#  Copyright 2015-2019 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os
import unittest

from swimai.structures import Record, Num, Attr, Slot, Text, RecordMap, Bool, Item, Extant, Absent, Value, \
    RecordFlags, RecordMapView, ValueBuilder
from test.utils import CustomString, CustomItem


class TestStructs(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        file_location = os.path.dirname(__file__)
        with open(os.path.join(file_location, 'expected_strings.txt')) as golden_file:
            self.expected_strings = dict()

            for line in golden_file:
                (key, val) = line.strip().split('=')
                self.expected_strings[key] = val

    def get_name(self):
        return self.id().split('.')[-1]

    def test_item_concat_single_num(self):
        # Given
        headers = Record.create().add_slot('node', 'foo').add_slot('lane', 'bar')
        body = Num.create_from(42)
        # When
        actual = Attr.create_attr('tag', headers).concat(body)
        # Then
        self.assertEqual(2, actual.size)
        self.assertIsInstance(actual.get_items()[0], Attr)
        self.assertIsInstance(actual.get_items()[1], Num)
        self.assertEqual(42, actual.get_items()[1].value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_concat_single_record(self):
        # Given
        headers = Record.create().add_slot('node', 'foo').add_slot('lane', 'bar')
        body = Record.create().add_slot('attr', 'hello').add_slot('message', 'world')
        # When
        actual = Attr.create_attr('tag', headers).concat(body)
        # Then
        self.assertEqual(3, actual.size)
        self.assertIsInstance(actual.get_items()[0], Attr)
        self.assertIsInstance(actual.get_items()[1], Slot)
        self.assertIsInstance(actual.get_items()[2], Slot)
        self.assertEqual('attr', actual.get_items()[1].key.value)
        self.assertEqual('hello', actual.get_items()[1].value.value)
        self.assertEqual('message', actual.get_items()[2].key.value)
        self.assertEqual('world', actual.get_items()[2].value.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_concat_multiple_text(self):
        # Given
        headers = Record.create().add_slot('node', 'foo').add_slot('lane', 'bar')
        headers.add(Num.create_from(42))
        body = Text.create_from('Polly the parrot')
        # When
        actual = headers.concat(body)
        # Then
        self.assertEqual(2, actual.size)
        self.assertIsInstance(actual.get_items()[0], RecordMap)
        self.assertEqual(3, actual.get_items()[0].size)
        self.assertIsInstance(actual.get_items()[1], Text)
        self.assertEqual('Polly the parrot', actual.get_items()[1].value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_concat_multiple_record(self):
        # Given
        headers = Record.create().add_slot('node', 'foo').add_slot('lane', 'bar')
        headers.add(Record.create().add_slot('node', 'boo').add_slot('lane', 'far'))
        body = Record.create().add_slot('node', 'poo').add_slot('lane', 'car')
        # When
        actual = headers.concat(body)
        # Then
        self.assertEqual(3, actual.size)
        self.assertIsInstance(actual.get_items()[0], RecordMap)
        self.assertEqual(3, actual.get_items()[0].size)
        self.assertIsInstance(actual.get_items()[1], Slot)
        self.assertEqual('node', actual.get_items()[1].key.value)
        self.assertEqual('poo', actual.get_items()[1].value.value)
        self.assertEqual('lane', actual.get_items()[2].key.value)
        self.assertEqual('car', actual.get_items()[2].value.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_concat_zero_bool(self):
        # Given
        headers = Record.create()
        body = Bool.create_from(True)
        # When
        actual = headers.concat(body)
        # Then
        self.assertEqual(2, actual.size)
        self.assertIsInstance(actual.get_items()[0], RecordMap)
        self.assertEqual(0, actual.get_items()[0].size)
        self.assertIsInstance(actual.get_items()[1], Bool)
        self.assertTrue(actual.get_items()[1].value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_concat_zero_record(self):
        # Given
        headers = Record.create()
        body = Record.create().add_slot('node', 'moo').add_slot('lane', 'cow')
        # When
        actual = headers.concat(body)
        # Then
        self.assertEqual(3, actual.size)
        self.assertIsInstance(actual.get_items()[0], RecordMap)
        self.assertEqual(0, actual.get_items()[0].size)
        self.assertIsInstance(actual.get_items()[1], Slot)
        self.assertIsInstance(actual.get_items()[2], Slot)
        self.assertEqual('node', actual.get_items()[1].key.value)
        self.assertEqual('moo', actual.get_items()[1].value.value)
        self.assertEqual('lane', actual.get_items()[2].key.value)
        self.assertEqual('cow', actual.get_items()[2].value.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_extant(self):
        # Given
        actual = Item.extant()
        # Then
        self.assertEqual(Extant.get_extant(), actual)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_absent(self):
        # Given
        actual = Item.absent()
        # Then
        self.assertEqual(Absent.get_absent(), actual)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_create_from_item(self):
        # Given
        record = Record.create().add_slot('node', 'boo').add_slot('lane', 'ghost')
        # When
        actual = Item.create_from(record)
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual(2, actual.size)
        self.assertEqual('node', actual.get_items()[0].key.value)
        self.assertEqual('boo', actual.get_items()[0].value.value)
        self.assertEqual('lane', actual.get_items()[1].key.value)
        self.assertEqual('ghost', actual.get_items()[1].value.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_create_from_tuple(self):
        # Given
        record = {'node': 'boo'}
        # When
        dict()
        actual = Item.create_from(record)
        # Then
        self.assertIsInstance(actual, Slot)
        self.assertEqual('node', actual.key)
        self.assertEqual('boo', actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_create_from_other(self):
        # Given
        record = 'monkey'
        # When
        actual = Item.create_from(record)
        # Then
        self.assertIsInstance(actual, Text)
        self.assertEqual('monkey', actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_attr_from_text_key(self):
        # Given
        key = Text.create_from('Ghost')
        value = 'Boo'
        # When
        actual = Attr.create_attr(key, value)
        # Then
        self.assertIsInstance(actual.key, Text)
        self.assertEqual('Ghost', actual.key.value)
        self.assertEqual('Boo', actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_attr_from_string_key(self):
        # Given
        key = 'Foo'
        value = Text.create_from('Bar')
        # When
        actual = Attr.create_attr(key, value)
        # Then
        self.assertIsInstance(actual.key, Text)
        self.assertEqual('Foo', actual.key.value)
        self.assertIsInstance(actual.value, Text)
        self.assertEqual('Bar', actual.value.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_attr_from_invalid_key(self):
        # Given
        key = 34
        value = 'Bar'
        # When
        with self.assertRaises(TypeError) as error:
            Attr.create_attr(key, value)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Invalid key: 34', message)

    def test_create_attr_from_empty_key(self):
        # Given
        key = None
        value = 'Bar'
        # When
        with self.assertRaises(TypeError) as error:
            Attr.create_attr(key, value)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Empty key for attribute!', message)

    def test_create_attr_from_empty_value(self):
        # Given
        key = 'Foo'
        value = None
        # When
        with self.assertRaises(TypeError) as error:
            Attr.create_attr(key, value)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Empty value for attribute!', message)

    def test_key_equals_str_true(self):
        # Given
        attribute = Attr.create_attr('Foo', 'Bar')
        # When
        actual = attribute.key_equals('Foo')
        # Then
        self.assertTrue(actual)

    def test_key_equals_str_false(self):
        # Given
        attribute = Attr.create_attr('Foo', 'Bar')
        # When
        actual = attribute.key_equals('Moo')
        # Then
        self.assertFalse(actual)

    def test_key_equals_attr_true(self):
        # Given
        attribute = Attr.create_attr('Hello', 'World')
        item = Attr.create_attr('Hello', 'Friend')
        # When
        actual = attribute.key_equals(item)
        # Then
        self.assertTrue(actual)

    def test_key_equals_attr_false(self):
        # Given
        attribute = Attr.create_attr('Hello', 'World')
        item = Attr.create_attr('Goodbye', 'Friend')
        # When
        actual = attribute.key_equals(item)
        # Then
        self.assertFalse(actual)

    def test_key_equals_other_false(self):
        # Given
        attribute = Attr.create_attr('Hello', 'World')
        item = CustomString('Hello')
        # When
        actual = attribute.key_equals(item)
        # Then
        self.assertTrue(actual)

    def test_key_equals_other_true(self):
        # Given
        attribute = Attr.create_attr('Hello', 'World')
        item = CustomString('Goodbye')
        # When
        actual = attribute.key_equals(item)
        # Then
        self.assertFalse(actual)

    def test_empty_value(self):
        # Given
        actual = Value()
        # Then
        self.assertEqual(Absent.get_absent(), actual.key)
        self.assertEqual(Absent.get_absent(), actual.value)
        self.assertEqual(0, actual.size)

    def test_create_value_from_object_none(self):
        # Given
        obj = None
        # When
        actual = Value.create_from(obj)
        # Then
        self.assertEqual(Extant.get_extant(), actual)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_value_from_object_value(self):
        # Given
        obj = Text.create_from('Boom')
        # When
        actual = Value.create_from(obj)
        # Then
        self.assertIsInstance(actual, Value)
        self.assertEqual('Boom', actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_value_from_object_str(self):
        # Given
        obj = 'Dog'
        # When
        actual = Value.create_from(obj)
        # Then
        self.assertIsInstance(actual, Text)
        self.assertEqual('Dog', actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_value_from_object_int(self):
        # Given
        obj = 33
        # When
        actual = Value.create_from(obj)
        # Then
        self.assertIsInstance(actual, Num)
        self.assertEqual(33, actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_value_from_object_float(self):
        # Given
        obj = 3.14
        # When
        actual = Value.create_from(obj)
        # Then
        self.assertIsInstance(actual, Num)
        self.assertEqual(3.14, actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_value_from_object_boolean(self):
        # Given
        obj = True
        # When
        actual = Value.create_from(obj)
        # Then
        self.assertIsInstance(actual, Bool)
        self.assertTrue(actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_value_from_object_invalid(self):
        # Given
        obj = CustomString('Foo')
        # When
        with self.assertRaises(TypeError) as error:
            Value.create_from(obj)
        # Then
        message = error.exception.args[0]
        self.assertEqual('CustomString(Foo) cannot be converted to Value!', message)

    def test_create_text_from_string(self):
        # Given
        string = 'Tea'
        # When
        actual = Text.create_from(string)
        # Then
        self.assertIsInstance(actual, Text)
        self.assertEqual('Tea', actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_text_from_empty_string(self):
        # Given
        string = ''
        # When
        actual = Text.create_from(string)
        # Then
        self.assertIsInstance(actual, Text)
        self.assertEqual('', actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_text_get_empty_first_time(self):
        # When
        actual = Text.get_empty()
        # Then
        self.assertEqual('', actual.value)
        self.assertEqual(Text.empty, actual)

    def test_text_get_empty_multiple_times(self):
        # Given
        Text.get_empty()
        # When
        actual = Text.get_empty()
        # Then
        self.assertEqual('', actual.value)
        self.assertEqual(Text.empty, actual)

    def test_text_string_value(self):
        # Given
        text = Text.create_from('Dog')
        # When
        actual = text.get_string_value()
        # Then
        self.assertIsInstance(actual, str)
        self.assertEqual('Dog', actual)

    def test_num_integer(self):
        # When
        num = Num(33)
        # Then
        self.assertEqual(33, num.value)
        self.assertEqual(33, num.get_num_value())
        self.assertEqual(self.expected_strings.get(self.get_name()), str(num))

    def test_num_float(self):
        # When
        num = Num(0.11)
        # Then
        self.assertEqual(0.11, num.value)
        self.assertEqual(0.11, num.get_num_value())
        self.assertEqual(self.expected_strings.get(self.get_name()), str(num))

    def test_create_num_from_positive_integer(self):
        # Given
        num = 99
        # When
        actual = Num.create_from(num)
        # Then
        self.assertIsInstance(actual, Num)
        self.assertEqual(99, actual.get_num_value())
        self.assertEqual(self.expected_strings.get(self.get_name()), str(num))

    def test_create_num_from_negative_integer(self):
        # Given
        num = -99
        # When
        actual = Num.create_from(num)
        # Then
        self.assertIsInstance(actual, Num)
        self.assertEqual(-99, actual.get_num_value())
        self.assertEqual(self.expected_strings.get(self.get_name()), str(num))

    def test_create_num_from_positive_float(self):
        # Given
        num = 9.9
        # When
        actual = Num.create_from(num)
        # Then
        self.assertIsInstance(actual, Num)
        self.assertEqual(9.9, actual.get_num_value())
        self.assertEqual(self.expected_strings.get(self.get_name()), str(num))

    def test_create_num_from_negative_float(self):
        # Given
        num = -9.9
        # When
        actual = Num.create_from(num)
        # Then
        self.assertIsInstance(actual, Num)
        self.assertEqual(-9.9, actual.get_num_value())
        self.assertEqual(self.expected_strings.get(self.get_name()), str(num))

    def test_bool_true(self):
        # Given
        boolean = Bool(True)
        # Then
        self.assertTrue(boolean.value)
        self.assertTrue(boolean.get_bool_value())
        self.assertEqual(self.expected_strings.get(self.get_name()), str(boolean))

    def test_bool_false(self):
        # Given
        boolean = Bool(False)
        # Then
        self.assertFalse(boolean.value)
        self.assertFalse(boolean.get_bool_value())
        self.assertEqual(self.expected_strings.get(self.get_name()), str(boolean))

    def test_create_bool_from_true_once(self):
        # Given
        boolean = True
        # When
        actual = Bool.create_from(boolean)
        # Then
        self.assertIsInstance(actual, Bool)
        self.assertEqual(Bool.TRUE, actual)
        self.assertTrue(actual.get_bool_value())

    def test_create_bool_from_true_twice(self):
        # Given
        boolean = True
        Bool.create_from(boolean)
        # When
        actual = Bool.create_from(boolean)
        # Then
        self.assertIsInstance(actual, Bool)
        self.assertEqual(Bool.TRUE, actual)
        self.assertTrue(actual.get_bool_value())

    def test_create_bool_from_false_once(self):
        # Given
        boolean = False
        # When
        actual = Bool.create_from(boolean)
        # Then
        self.assertIsInstance(actual, Bool)
        self.assertEqual(Bool.FALSE, actual)
        self.assertFalse(actual.get_bool_value())

    def test_create_bool_from_false_twice(self):
        # Given
        boolean = False
        Bool.create_from(boolean)
        # When
        actual = Bool.create_from(boolean)
        # Then
        self.assertIsInstance(actual, Bool)
        self.assertEqual(Bool.FALSE, actual)
        self.assertFalse(actual.get_bool_value())

    def test_get_absent_once(self):
        # When
        actual = Absent.get_absent()
        # Then
        self.assertIsInstance(actual, Absent)
        self.assertEqual(Absent.absent, actual)
        self.assertEqual(Absent.get_absent(), actual)

    def test_get_absent_twice(self):
        # Given
        Absent.get_absent()
        # When
        actual = Absent.get_absent()
        # Then
        self.assertIsInstance(actual, Absent)
        self.assertEqual(Absent.absent, actual)
        self.assertEqual(Absent.get_absent(), actual)

    def test_get_extant_once(self):
        # When
        actual = Extant.get_extant()
        # Then
        self.assertIsInstance(actual, Extant)
        self.assertEqual(Extant.extant, actual)
        self.assertEqual(Extant.get_extant(), actual)

    def test_get_extant_twice(self):
        # Given
        Extant.get_extant()
        # When
        actual = Extant.get_extant()
        # Then
        self.assertIsInstance(actual, Extant)
        self.assertEqual(Extant.extant, actual)
        self.assertEqual(Extant.get_extant(), actual)

    def test_slot_key_value(self):
        # Given
        slot = Slot('Foo', 'Bar')
        # Then
        self.assertIsInstance(slot, Slot)
        self.assertEqual('Foo', slot.key)
        self.assertEqual('Bar', slot.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(slot))

    def test_slot_key_only(self):
        # Given
        slot = Slot('Foo')
        # Then
        self.assertIsInstance(slot, Slot)
        self.assertEqual('Foo', slot.key)
        self.assertEqual(Extant.get_extant(), slot.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(slot))

    def test_create_slot_key_value(self):
        # Given
        key = 'dog'
        value = 'bark'
        # When
        actual = Slot.create_slot(key, value)
        # Then
        self.assertIsInstance(actual, Slot)
        self.assertEqual('dog', actual.key)
        self.assertEqual('bark', actual.value)

    def test_create_slot_key_only(self):
        # Given
        key = 'dog'
        # When
        actual = Slot.create_slot(key)
        # Then
        self.assertIsInstance(actual, Slot)
        self.assertEqual('dog', actual.key)
        self.assertEqual(Extant.get_extant(), actual.value)

    def test_create_slot_no_key(self):
        # Given
        key = None
        # When
        with self.assertRaises(TypeError) as error:
            Slot.create_slot(key)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Empty key for slot!', message)

    def test_record_create(self):
        # When
        actual = Record.create()
        # Then
        self.assertIsInstance(actual, RecordMap)

    def test_record_add_all_single(self):
        # Given
        actual = Record.create()
        item = Text.create_from('Foo')
        items = [item]
        # When
        actual_response = actual.add_all(items)
        # Then
        self.assertEqual(1, actual.size)
        self.assertEqual(1, actual.item_count)
        self.assertEqual(item, actual.get_item(0))
        self.assertTrue(actual_response)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_record_add_all_multiple(self):
        # Given
        actual = Record.create()
        first_item = Text.create_from('Foo')
        second_item = Text.create_from('Bar')
        third_item = Text.create_from('Baz')
        items = [first_item, second_item, third_item]
        # When
        actual_response = actual.add_all(items)
        # Then
        self.assertEqual(3, actual.size)
        self.assertEqual(3, actual.item_count)
        self.assertEqual(first_item, actual.get_item(0))
        self.assertEqual(second_item, actual.get_item(1))
        self.assertEqual(third_item, actual.get_item(2))
        self.assertTrue(actual_response)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_record_add_all_empty(self):
        # Given
        actual = Record.create()
        items = []
        # When
        actual_response = actual.add_all(items)
        # Then
        self.assertEqual(0, actual.size)
        self.assertEqual(0, actual.item_count)
        self.assertEqual(False, actual_response)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_record_add_slot_key_item_value_item(self):
        # Given
        actual = Record.create()
        key = Text.create_from('Good')
        value = Text.create_from('Dog')
        # When
        actual = actual.add_slot(key, value)
        # Then
        self.assertEqual(1, actual.size)
        self.assertIsInstance(actual, Record)
        self.assertIsInstance(actual.get_item(0), Slot)
        self.assertEqual(key, actual.get_item(0).key)
        self.assertEqual(value, actual.get_item(0).value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_record_add_slot_key_string_value_item(self):
        # Given
        actual = Record.create()
        key = 'Baz'
        value = Text.create_from('Qux')
        # When
        actual = actual.add_slot(key, value)
        # Then
        self.assertEqual(1, actual.size)
        self.assertIsInstance(actual, Record)
        self.assertIsInstance(actual.get_item(0), Slot)
        self.assertEqual('Baz', actual.get_item(0).key.value)
        self.assertEqual('Qux', actual.get_item(0).value.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_record_add_slot_key_item_value_string(self):
        # Given
        actual = Record.create()
        key = Text.create_from('Foo')
        value = 'Bar'
        # When
        actual = actual.add_slot(key, value)
        # Then
        self.assertEqual(1, actual.size)
        self.assertIsInstance(actual, Record)
        self.assertIsInstance(actual.get_item(0), Slot)
        self.assertEqual('Foo', actual.get_item(0).key.value)
        self.assertEqual('Bar', actual.get_item(0).value.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_record_add_slot_key_string_value_string(self):
        # Given
        actual = Record.create()
        key = 'Wibble'
        value = 'Wobble'
        # When
        actual = actual.add_slot(key, value)
        # Then
        self.assertEqual(1, actual.size)
        self.assertIsInstance(actual, Record)
        self.assertIsInstance(actual.get_item(0), Slot)
        self.assertEqual('Wibble', actual.get_item(0).key.value)
        self.assertEqual('Wobble', actual.get_item(0).value.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_record_get_headers_record(self):
        # Given
        record = Record.create()
        body = Record.create()
        body.add_slot('Foo', 'Bar')
        record.add(Attr.create_attr(Text.create_from('cat'), body))
        # When
        actual = record.get_headers('cat')
        # Then
        self.assertIsInstance(actual, Record)
        self.assertEqual(1, actual.size)
        self.assertEqual('Foo', actual.get_item(0).key.value)
        self.assertEqual('Bar', actual.get_item(0).value.value)

    def test_record_get_headers_object(self):
        # Given
        record = Record.create()
        body = Slot.create_slot('Baz', 'Qux')
        record.add(Attr.create_attr(Text.create_from('dog'), body))
        # When
        actual = record.get_headers('dog')
        # Then
        self.assertIsInstance(actual, Record)
        self.assertEqual(1, actual.size)
        self.assertEqual('Baz', actual.get_item(0).key)
        self.assertEqual('Qux', actual.get_item(0).value)

    def test_record_get_headers_none(self):
        # Given
        record = Record.create()
        body = Slot.create_slot('Hello', 'World')
        record.add(Attr.create_attr(Text.create_from('mouse'), body))
        # When
        actual = record.get_headers('bird')
        # Then
        self.assertIsNone(actual)

    def test_record_get_head(self):
        # Given
        record = Record.create()
        body = Slot.create_slot('Hello', 'World')
        first_item = Attr.create_attr(Text.create_from('bird'), body)
        second_item = Attr.create_attr(Text.create_from('frog'), body)
        record.add(first_item)
        record.add(second_item)
        # When
        actual = record.get_head()
        # Then
        self.assertEqual(first_item, actual)
        self.assertNotEqual(second_item, actual)

    def test_record_bind(self):
        # Given
        record = Record.create()
        body = Slot.create_slot('Hello', 'Friend')
        first_item = Attr.create_attr(Text.create_from('dog'), body)
        record.add(first_item)
        # When
        actual = record.bind()
        # Then
        self.assertEqual(record, actual)

    def test_record_map_size_empty(self):
        # Given
        record = RecordMap.create()
        # When
        actual = record.size
        # Then
        self.assertEqual(0, actual)

    def test_record_map_size_one(self):
        # Given
        record = RecordMap.create_record_map(Text.create_from('Test'))
        # When
        actual = record.size
        # Then
        self.assertEqual(1, actual)

    def test_record_map_size_three(self):
        # Given
        record = RecordMap.create()
        record.add(Text.create_from('This'))
        record.add(Text.create_from('is'))
        record.add(Text.create_from('test'))
        # When
        actual = record.size
        # Then
        self.assertEqual(3, actual)

    def test_record_map_tag_attr(self):
        # Given
        record = RecordMap.create()
        record.add(Attr.create_attr('Foo', 'Bar'))
        # When
        actual = record.tag
        # Then
        self.assertEqual('Foo', actual)

    def test_record_map_tag_slot(self):
        # Given
        record = RecordMap.create()
        record.add(Slot.create_slot('Foo', 'Bar'))
        # When
        actual = record.tag
        # Then
        self.assertEqual(None, actual)

    def test_record_map_tag_empty(self):
        # Given
        record = RecordMap.create()
        # When
        actual = record.tag
        # Then
        self.assertEqual(None, actual)

    def test_record_map_create(self):
        # When
        actual = RecordMap.create()
        # Then
        self.assertEqual(0, actual.size)
        self.assertEqual(RecordFlags.ALIASED.value, actual.flags)

    def test_record_map_create_with_empty(self):
        # Given
        item = None
        # When
        actual = RecordMap.create_record_map(item)
        # Then
        self.assertEqual(1, actual.size)
        self.assertEqual(0, actual.field_count)
        self.assertEqual(Extant.get_extant(), actual.get_item(0))

    def test_record_map_create_with_field(self):
        # Given
        item = Attr.create_attr('Foo', 'Bar')
        # When
        actual = RecordMap.create_record_map(item)
        # Then
        self.assertEqual(1, actual.size)
        self.assertEqual(1, actual.field_count)
        self.assertEqual(item, actual.get_item(0))
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_record_map_create_with_value(self):
        # Given
        item = Value.create_from('Baz')
        # When
        actual = RecordMap.create_record_map(item)
        # Then
        self.assertEqual(1, actual.size)
        self.assertEqual(0, actual.field_count)
        self.assertEqual(item, actual.get_item(0))

    def test_record_map_get_item_existing(self):
        # Given
        record_map = RecordMap.create()
        first_item = Attr.create_attr('First', '1st')
        second_item = Attr.create_attr('Second', '2nd')
        third_item = Attr.create_attr('Third', '3rd')
        record_map.add(first_item)
        record_map.add(second_item)
        record_map.add(third_item)
        # When
        actual = record_map.get_item(1)
        # Then
        self.assertEqual(second_item, actual)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_record_map_get_item_underflow(self):
        # Given
        record_map = RecordMap.create()
        first_item = Attr.create_attr('First', '1st')
        second_item = Attr.create_attr('Second', '2nd')
        record_map.add(first_item)
        record_map.add(second_item)
        # When
        actual = record_map.get_item(-1)
        # Then
        self.assertEqual(Absent.get_absent(), actual)

    def test_record_map_get_item_overflow(self):
        # Given
        record_map = RecordMap.create()
        first_item = Attr.create_attr('First', '1st')
        record_map.add(first_item)
        # When
        actual = record_map.get_item(2)
        # Then
        self.assertEqual(Absent.get_absent(), actual)

    def test_record_map_get_items_empty(self):
        # Given
        record_map = RecordMap.create()
        # When
        actual = record_map.get_items()
        # Then
        self.assertEqual([], actual)

    def test_record_map_get_items_single(self):
        # Given
        record_map = RecordMap.create()
        first_item = Attr.create_attr('Hello', 'World')
        record_map.add(first_item)
        # When
        actual = record_map.get_items()
        # Then
        self.assertEqual([first_item], actual)

    def test_record_map_get_items_multiple(self):
        # Given
        record_map = RecordMap.create()
        first_item = Attr.create_attr('Hello', 'World')
        second_item = Attr.create_attr('Bye', 'World')
        record_map.add(first_item)
        record_map.add(second_item)
        # When
        actual = record_map.get_items()
        # Then
        self.assertEqual([first_item, second_item], actual)

    def test_record_map_get_body_empty_without_head(self):
        # Given
        record_map = RecordMap.create()
        # When
        actual = record_map.get_body()
        # Then
        self.assertEqual(Absent.get_absent(), actual)

    def test_record_map_get_body_empty_with_head(self):
        # Given
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr('Hello', 'World'))
        # When
        actual = record_map.get_body()
        # Then
        self.assertEqual(Absent.get_absent(), actual)

    def test_record_map_get_body_single_value(self):
        # Given
        record_map = RecordMap.create()
        body = Text.create_from('Moo')
        record_map.add(Attr.create_attr('Hello', 'World'))
        record_map.add(body)
        # When
        actual = record_map.get_body()
        # Then
        self.assertEqual(body, actual)

    def test_record_map_get_body_single_object(self):
        # Given
        record_map = RecordMap.create()
        body = Attr.create_attr('Moo', 'Moo')
        record_map.add(Attr.create_attr('Hello', 'World'))
        record_map.add(body)
        # When
        actual = record_map.get_body()
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual(body, actual.get_item(0))

    def test_record_map_get_body_multiple_(self):
        # Given
        record_map = RecordMap.create()
        first = Attr.create_attr('Moo', 'Moo')
        second = Attr.create_attr('Boo', 'Boo')
        record_map.add(Attr.create_attr('Goodbye', 'World'))
        record_map.add(first)
        record_map.add(second)
        # When
        actual = record_map.get_body()
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual(first, actual.get_item(0))
        self.assertEqual(second, actual.get_item(1))

    def test_record_immutable_map_add(self):
        # Given
        record_map = RecordMap.create()
        record_map.flags = RecordFlags.IMMUTABLE.value
        item = Attr.create_attr('Moo', 'Moo')
        # When
        with self.assertRaises(TypeError) as error:
            record_map.add(item)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Cannot add item to immutable record!', message)

    def test_record_aliased_map_add(self):
        # Given
        record_map = RecordMap.create()
        record_map.flags = RecordFlags.ALIASED.value
        item = Attr.create_attr('Moo', 'Moo')
        # When
        actual_response = record_map.add(item)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual([item], record_map.items)
        self.assertEqual(1, record_map.item_count)
        self.assertEqual(1, record_map.field_count)
        self.assertEqual(0, record_map.flags)
        self.assertEqual(None, record_map.fields)

    def test_record_mutable_map_add(self):
        # Given
        record_map = RecordMap.create()
        record_map.flags = 0
        item = Attr.create_attr('Foo', 'Foo')
        # When
        actual_response = record_map.add(item)
        # Then
        self.assertEqual([item], record_map.items)
        self.assertTrue(actual_response)
        self.assertEqual(1, record_map.item_count)
        self.assertEqual(1, record_map.field_count)

    def test_record_map_add_mutable_non_field(self):
        # Given
        record_map = RecordMap.create()
        record_map.flags = 0
        item = Value.create_from('Foo')
        # When
        actual_response = record_map.add(item)
        # Then
        self.assertEqual([item], record_map.items)
        self.assertTrue(actual_response)
        self.assertEqual(1, record_map.item_count)
        self.assertEqual(0, record_map.field_count)

    def test_record_map_add_mutable_field_with_hashtable_unique(self):
        # Given
        record_map = RecordMap.create()
        record_map.flags = 0
        first_item = Attr.create_attr('Foo', 'Foo')
        second_item = Attr.create_attr('Moo', 'Moo')

        # When
        record_map.add(first_item)
        record_map.contains_key('Foo')
        record_map.add(second_item)
        # Then
        self.assertEqual([first_item, second_item], record_map.items)
        self.assertEqual(2, record_map.item_count)
        self.assertEqual(2, record_map.field_count)
        self.assertEqual({'Foo': first_item, 'Moo': second_item}, record_map.fields)

    def test_record_map_add_mutable_field_with_hashtable_duplicate(self):
        # Given
        record_map = RecordMap.create()
        record_map.flags = 0
        first_item = Attr.create_attr('Boo', 'Foo')
        second_item = Attr.create_attr('Boo', 'Moo')
        # When
        record_map.add(first_item)
        record_map.add(second_item)
        record_map.contains_key('Boo')
        # Then
        self.assertEqual([first_item, second_item], record_map.items)
        self.assertEqual(2, record_map.item_count)
        self.assertEqual(2, record_map.field_count)
        self.assertEqual({'Boo': second_item}, record_map.fields)

    def test_record_map_add_aliased_non_field(self):
        # Given
        record_map = RecordMap.create()
        record_map.flags = 2
        item = Value.create_from('Boo')
        # When
        actual_response = record_map.add(item)
        # Then
        self.assertEqual([item], record_map.items)
        self.assertEqual(True, actual_response)
        self.assertEqual(1, record_map.item_count)
        self.assertEqual(0, record_map.field_count)

    def test_record_map_commit_mutable(self):
        # Given
        record_map = RecordMap.create()
        record_map.flags = 0
        # When
        actual = record_map.commit()
        # Then
        self.assertEqual(record_map, actual)
        self.assertEqual(RecordFlags.IMMUTABLE.value, actual.flags)

    def test_record_map_commit_aliased(self):
        # Given
        record_map = RecordMap.create()
        record_map.flags = 2
        # When
        actual = record_map.commit()
        # Then
        self.assertEqual(record_map, actual)
        self.assertEqual(3, actual.flags)

    def test_record_map_commit_immutable(self):
        # Given
        record_map = RecordMap.create()
        record_map.flags = 1
        # When
        actual = record_map.commit()
        # Then
        self.assertEqual(record_map, actual)
        self.assertEqual(1, actual.flags)

    def test_record_map_contains_key_value(self):
        # Given
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr('Foo', 'Poo'))
        record_map.add(Attr.create_attr('Moo', 'Boo'))
        # When
        actual = record_map.contains_key(Text.create_from('Moo'))
        # Then
        self.assertTrue(actual)

    def test_record_map_contains_key_string(self):
        # Given
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr('Foo', 'Poo'))
        record_map.add(Attr.create_attr('Moo', 'Boo'))
        # When
        actual = record_map.contains_key('Foo')
        # Then
        self.assertTrue(actual)

    def test_record_map_contains_key_duplicate(self):
        # Given
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr('Foo', 'Poo'))
        record_map.add(Attr.create_attr('Foo', 'Boo'))
        # When
        actual = record_map.contains_key('Foo')
        # Then
        self.assertTrue(actual)

    def test_record_map_contains_key_false(self):
        # Given
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr('Foo', 'Poo'))
        record_map.add(Attr.create_attr('Moo', 'Boo'))
        # When
        actual = record_map.contains_key('Poo')
        # Then
        self.assertFalse(actual)

    def test_record_map_contains_key_field_count_empty(self):
        # Given
        record_map = RecordMap.create()
        # When
        actual = record_map.contains_key('Poo')
        # Then
        self.assertFalse(actual)

    def test_record_map_branch_empty(self):
        # Given
        original_record = Record.create()
        original_record.flags = 0
        # When
        copy_record = original_record.branch()
        # Then
        self.assertEqual(original_record.items, copy_record.items)
        self.assertEqual(original_record.fields, copy_record.fields)
        self.assertEqual(original_record.item_count, copy_record.item_count)
        self.assertEqual(original_record.field_count, copy_record.field_count)
        self.assertEqual(2, original_record.flags)
        self.assertEqual(original_record.flags, copy_record.flags)

    def test_record_map_branch_single(self):
        # Given
        original_record = Record.create()
        original_record.flags = 0
        original_record.add(Attr.create_attr('Foo', 'Bar'))
        # When
        copy_record = original_record.branch()
        # Then
        self.assertEqual(2, original_record.flags)
        self.assertEqual(original_record.items, copy_record.items)
        self.assertEqual(original_record.fields, copy_record.fields)
        self.assertEqual(1, copy_record.item_count)
        self.assertEqual(1, copy_record.field_count)

    def test_record_map_branch_multiple(self):
        # Given
        original_record = Record.create()
        original_record.flags = 0
        original_record.add(Attr.create_attr('Foo', 'Bar'))
        original_record.add(Attr.create_attr('Baz', 'Qux'))
        # When
        copy_record = original_record.branch()
        # Then
        self.assertEqual(2, original_record.flags)
        self.assertEqual(original_record.items, copy_record.items)
        self.assertEqual(original_record.fields, copy_record.fields)
        self.assertEqual(2, copy_record.item_count)
        self.assertEqual(2, copy_record.field_count)

    def test_record_map_branch_with_hashtable(self):
        # Given
        original_record = Record.create()
        original_record.flags = 0
        original_record.add(Attr.create_attr('Bar', 'Baz'))
        original_record.contains_key('Bar')
        # When
        copy_record = original_record.branch()
        # Then
        self.assertIsNotNone(copy_record.fields)
        self.assertEqual(original_record.items, copy_record.items)
        self.assertEqual(original_record.fields, copy_record.fields)
        self.assertEqual(1, copy_record.item_count)
        self.assertEqual(1, copy_record.field_count)

    def test_record_map_branch_mutate_original(self):
        # Given
        original_record = Record.create()
        original_record.add(Attr.create_attr('K', 'V'))
        original_record.add(Attr.create_attr('A', 'B'))
        # When
        original_record.contains_key('K')
        copy_record = original_record.branch()
        original_record.add(Attr.create_attr(Text.create_from('P'), Text.create_from('V')))
        original_record.contains_key('K')
        # Then
        self.assertEqual(0, original_record.flags)
        self.assertEqual(2, copy_record.flags)
        self.assertNotEqual(original_record.items, copy_record.items)
        self.assertEqual(3, original_record.item_count)
        self.assertEqual(2, copy_record.item_count)
        self.assertEqual(3, original_record.field_count)
        self.assertEqual(2, copy_record.field_count)
        self.assertIsNotNone(original_record.fields)
        self.assertIsNotNone(copy_record.fields)
        self.assertEqual(3, len(original_record.fields))
        self.assertEqual(2, len(copy_record.fields))

    def test_record_map_branch_mutate_copy(self):
        # Given
        original_record = Record.create()
        original_record.add(Attr.create_attr('F', 'L'))
        original_record.add(Attr.create_attr('M', 'S'))
        # When
        original_record.contains_key('P')
        copy_record = original_record.branch()
        copy_record.add(Attr.create_attr(Text.create_from('P'), Text.create_from('V')))
        copy_record.add(Attr.create_attr(Text.create_from('T'), Text.create_from('B')))
        copy_record.contains_key('H')
        # Then
        self.assertEqual(2, original_record.flags)
        self.assertEqual(0, copy_record.flags)
        self.assertNotEqual(original_record.items, copy_record.items)
        self.assertEqual(2, original_record.item_count)
        self.assertEqual(4, copy_record.item_count)
        self.assertEqual(2, original_record.field_count)
        self.assertEqual(4, copy_record.field_count)
        self.assertIsNotNone(original_record.fields)
        self.assertIsNotNone(copy_record.fields)
        self.assertEqual(2, len(original_record.fields))
        self.assertEqual(4, len(copy_record.fields))

    def test_record_map_view(self):
        # Given
        record = Record.create()
        record.add(Attr.create_attr('A', 'B'))
        record.add(Attr.create_attr('C', 'D'))
        record.add(Attr.create_attr('E', 'F'))
        record.add(Attr.create_attr('G', 'H'))
        # When
        actual = RecordMapView(record, 1, 3)
        # Then
        self.assertEqual(2, actual.size)
        self.assertEqual(1, actual.lower)
        self.assertEqual(3, actual.upper)

    def test_record_map_view_get_item_existing(self):
        # Given
        record = Record.create()
        record.add(Attr.create_attr('Z', 'X'))
        record.add(Attr.create_attr('Y', 'W'))
        record.add(Attr.create_attr('P', 'L'))
        record.add(Attr.create_attr('M', 'N'))
        # When
        actual = RecordMapView(record, 1, 3)
        # Then
        self.assertEqual('W', actual.get_item(0).value)
        self.assertEqual('L', actual.get_item(1).value)

    def test_record_map_view_get_item_underflow(self):
        # Given
        record = Record.create()
        record.add(Attr.create_attr('G', 'V'))
        record.add(Attr.create_attr('H', 'B'))
        record.add(Attr.create_attr('J', 'N'))
        record.add(Attr.create_attr('K', 'M'))
        # When
        actual = RecordMapView(record, 1, 3)
        # Then
        self.assertEqual(Absent.get_absent(), actual.get_item(-1).value)

    def test_record_map_view_get_item_overflow(self):
        # Given
        record = Record.create()
        record.add(Attr.create_attr('F', 'V'))
        record.add(Attr.create_attr('Y', 'B'))
        record.add(Attr.create_attr('U', 'N'))
        record.add(Attr.create_attr('I', 'M'))
        # When
        actual = RecordMapView(record, 1, 3)
        # Then
        self.assertEqual(Absent.get_absent(), actual.get_item(3).value)

    def test_record_map_view_get_items_none(self):
        # Given
        record = Record.create()
        record.add(Attr.create_attr('U', 'N'))
        record.add(Attr.create_attr('I', 'M'))
        # When
        actual = RecordMapView(record, 4, 5)
        # Then
        self.assertEqual([], actual.get_items())

    def test_record_map_view_get_items_single(self):
        # Given
        record = Record.create()
        first = Attr.create_attr('U', 'N')
        second = Attr.create_attr('A', 'N')
        third = Attr.create_attr('I', 'M')
        record.add(first)
        record.add(second)
        record.add(third)
        # When
        actual = RecordMapView(record, 1, 2)
        # Then
        self.assertEqual([second], actual.get_items())

    def test_record_map_view_get_items_multiple(self):
        # Given
        record = Record.create()
        first = Attr.create_attr('A', 'K')
        second = Attr.create_attr('K', 'P')
        third = Attr.create_attr('Q', 'F')
        record.add(first)
        record.add(second)
        record.add(third)
        # When
        actual = RecordMapView(record, 1, 3)
        # Then
        self.assertEqual([second, third], actual.get_items())

    def test_record_map_view_immutable_add(self):
        # Given
        record = Record.create()
        record.add(Attr.create_attr('L', 'K'))
        record.add(Attr.create_attr('T', 'P'))
        record.add(Attr.create_attr('R', 'F'))
        new_item = Attr.create_attr('New', 'Item')
        record.flags = 1
        record_map_view = RecordMapView(record, 1, 3)
        # When
        with self.assertRaises(TypeError) as error:
            record_map_view.add(new_item, 2)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Cannot add item to immutable record!', message)

    def test_record_map_view_out_of_bounds_add(self):
        # Given
        record = Record.create()
        record.add(Attr.create_attr('T', 'K'))
        record.add(Attr.create_attr('I', 'P'))
        record.add(Attr.create_attr('M', 'F'))
        new_item = Attr.create_attr('Foo', 'Bar')
        record_map_view = RecordMapView(record, 1, 3)
        # When
        with self.assertRaises(IndexError) as error:
            record_map_view.add(new_item, 5)
        # Then
        message = error.exception.args[0]
        self.assertEqual(f'Index 5 is out of range!', message)

    def test_record_map_view_add_field(self):
        # Given
        record = Record.create()
        record.add(Attr.create_attr('T', 'K'))
        record.add(Attr.create_attr('C', 'P'))
        record.add(Attr.create_attr('M', 'F'))
        new_item = Attr.create_attr('Field', 'Item')
        record_map_view = RecordMapView(record, 1, 3)
        # When
        actual_response = record_map_view.add(new_item, 2)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual(3, record_map_view.size)
        self.assertEqual(4, record_map_view.upper)
        self.assertEqual(1, record_map_view.lower)
        self.assertEqual(new_item, record_map_view.get_item(2))
        self.assertEqual(4, record_map_view.record.item_count)
        self.assertEqual(4, record_map_view.record.field_count)
        self.assertEqual(0, record_map_view.record.flags)
        self.assertEqual(None, record_map_view.record.fields)

    def test_record_map_view_add_value(self):
        # Given
        record = Record.create()
        record.add(Attr.create_attr('T', 'K'))
        record.add(Attr.create_attr('T', 'P'))
        new_item = Text.create_from('Text')
        record_map_view = RecordMapView(record, 0, 2)
        # When
        actual_response = record_map_view.add(new_item, 1)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual(3, record_map_view.size)
        self.assertEqual(3, record_map_view.upper)
        self.assertEqual(0, record_map_view.lower)
        self.assertEqual(new_item, record_map_view.get_item(1))
        self.assertEqual(3, record_map_view.record.item_count)
        self.assertEqual(2, record_map_view.record.field_count)
        self.assertEqual(0, record_map_view.record.flags)
        self.assertEqual(None, record_map_view.record.fields)

    def test_record_map_view_add_none_index(self):
        # Given
        record = Record.create()
        record.add(Attr.create_attr('B', 'H'))
        record.add(Attr.create_attr('E', 'R'))
        new_item = Text.create_from('Foo')
        record_map_view = RecordMapView(record, 1, 2)
        # When
        actual_response = record_map_view.add(new_item)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual(2, record_map_view.size)
        self.assertEqual(3, record_map_view.upper)
        self.assertEqual(1, record_map_view.lower)
        self.assertEqual(new_item, record_map_view.get_item(1))

    def test_record_map_view_branch_empty(self):
        # Given
        record = Record.create()
        record.add(Attr.create_attr('Q', 'H'))
        record.add(Attr.create_attr('W', 'R'))
        record_map_view = RecordMapView(record, 0, 0)
        # When
        actual = record_map_view.branch()
        # Then
        self.assertEqual(0, actual.size)
        self.assertEqual(0, actual.field_count)
        self.assertIsInstance(actual, RecordMap)
        self.assertNotEqual(record_map_view.record.items, actual.items)
        self.assertEqual([], actual.items)
        self.assertEqual(None, actual.fields)

    def test_record_map_view_branch_single(self):
        # Given
        record = Record.create()
        first = Attr.create_attr('R', 'H')
        second = Attr.create_attr('Y', 'R')
        record.add(first)
        record.add(second)
        record_map_view = RecordMapView(record, 0, 1)
        # When
        actual = record_map_view.branch()
        # Then
        self.assertEqual(1, actual.size)
        self.assertEqual(1, actual.field_count)
        self.assertIsInstance(actual, RecordMap)
        self.assertNotEqual(record_map_view.record.items, actual.items)
        self.assertEqual([first], actual.items)
        self.assertEqual(None, actual.fields)

    def test_record_map_view_branch_multiple_fields(self):
        # Given
        record = Record.create()
        first = Attr.create_attr('T', 'H')
        second = Attr.create_attr('E', 'R')
        third = Attr.create_attr('P', 'Q')
        record.add(first)
        record.add(second)
        record.add(third)
        record_map_view = RecordMapView(record, 0, 2)
        # When
        actual = record_map_view.branch()
        # Then
        self.assertEqual(2, actual.size)
        self.assertEqual(2, actual.field_count)
        self.assertIsInstance(actual, RecordMap)
        self.assertNotEqual(record_map_view.record.items, actual.items)
        self.assertEqual([first, second], actual.items)
        self.assertEqual(None, actual.fields)

    def test_record_map_view_branch_multiple_values(self):
        # Given
        record = Record.create()
        second = Text.create_from('Bar')
        third = Text.create_from('Baz')
        fourth = Text.create_from('Qux')
        record.add(Text.create_from('Foo'))
        record.add(second)
        record.add(third)
        record.add(fourth)
        record_map_view = RecordMapView(record, 1, 4)
        # When
        actual = record_map_view.branch()
        # Then
        self.assertEqual(3, actual.size)
        self.assertEqual(0, actual.field_count)
        self.assertIsInstance(actual, RecordMap)
        self.assertNotEqual(record_map_view.record.items, actual.items)
        self.assertEqual([second, third, fourth], actual.items)
        self.assertEqual(None, actual.fields)

    def test_record_map_view_branch_multiple_values_and_fields(self):
        # Given
        record = Record.create()
        first = Text.create_from('Bar')
        second = Attr.create_attr('E', 'R')
        third = Text.create_from('Baz')
        fourth = Attr.create_attr('P', 'Q')
        record.add(first)
        record.add(second)
        record.add(third)
        record.add(fourth)
        record_map_view = RecordMapView(record, 0, 3)
        # When
        actual = record_map_view.branch()
        # Then
        self.assertEqual(3, actual.size)
        self.assertEqual(1, actual.field_count)
        self.assertIsInstance(actual, RecordMap)
        self.assertNotEqual(record_map_view.record.items, actual.items)
        self.assertEqual([first, second, third], actual.items)
        self.assertEqual(None, actual.fields)

    def test_value_builder_add_field_to_empty_record(self):
        # Given
        value_builder = ValueBuilder()
        item = Attr.create_attr('Foo', 'Boo')
        # When
        actual_response = value_builder.add(item)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual(1, value_builder.record.size)
        self.assertEqual(None, value_builder.value)

    def test_value_builder_add_field_existing_record(self):
        # Given
        value_builder = ValueBuilder()
        value_builder.add(Attr.create_attr('Baz', 'Qux'))
        item = Attr.create_attr('Foo', 'Boo')
        # When
        actual_response = value_builder.add(item)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual(2, value_builder.record.size)
        self.assertEqual(None, value_builder.value)

    def test_value_builder_add_field_to_empty_record_and_existing_value(self):
        # Given
        value_builder = ValueBuilder()
        value_builder.add(Text.create_from('Baz'))
        item = Attr.create_attr('Moo', 'Cow')
        # When
        actual_response = value_builder.add(item)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual(2, value_builder.record.size)
        self.assertEqual(None, value_builder.value)
        self.assertIsInstance(value_builder.record.get_item(0), Text)
        self.assertIsInstance(value_builder.record.get_item(1), Attr)

    def test_value_builder_add_value_to_existing_record(self):
        # Given
        value_builder = ValueBuilder()
        value_builder.add(Attr.create_attr('Moo', 'Cow'))
        value_builder.add(Attr.create_attr('Boo', 'Ghost'))
        item = Text.create_from('Baz')
        # When
        actual_response = value_builder.add(item)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual(3, value_builder.record.size)
        self.assertEqual(None, value_builder.value)
        self.assertIsInstance(value_builder.record.get_item(0), Attr)
        self.assertIsInstance(value_builder.record.get_item(1), Attr)
        self.assertIsInstance(value_builder.record.get_item(2), Text)

    def test_value_builder_add_value_to_empty_record_and_value(self):
        # Given
        value_builder = ValueBuilder()
        item = Text.create_from('Bar')
        # When
        actual_response = value_builder.add(item)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual(None, value_builder.record)
        self.assertEqual('Bar', value_builder.value.value)

    def test_value_builder_add_value_to_empty_record_exiting_value(self):
        # Given
        value_builder = ValueBuilder()
        value_builder.add(Text.create_from('Bar'))
        value_builder.add(Text.create_from('Boo'))
        value_builder.add(Text.create_from('Foo'))
        item = Text.create_from('Moo')
        # When
        actual_response = value_builder.add(item)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual(4, value_builder.record.size)
        self.assertEqual(None, value_builder.value)
        self.assertIsInstance(value_builder.record.get_item(0), Text)
        self.assertIsInstance(value_builder.record.get_item(1), Text)

    def test_value_builder_add_not_supported(self):
        # Given
        value_builder = ValueBuilder()
        item = CustomItem()
        # When
        with self.assertRaises(TypeError) as error:
            value_builder.add(item)
        # Then
        message = error.exception.args[0]
        self.assertEqual(f'Item of type CustomItem is not supported by the Value Builder', message)

    def test_value_builder_bind_with_record(self):
        # Given
        value_builder = ValueBuilder()
        attr = Attr.create_attr('Foo', 'Bar')
        value_builder.add(attr)
        # When
        actual = value_builder.bind()
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual(1, actual.size)
        self.assertEqual(attr, actual.get_item(0))

    def test_value_builder_bind_with_value(self):
        # Given
        value_builder = ValueBuilder()
        value = Text.create_from('Foo')
        value_builder.add(value)
        # When
        actual = value_builder.bind()
        # Then
        self.assertEqual(value, actual)

    def test_value_builder_bind_with_empty_record_and_value(self):
        # Given
        value_builder = ValueBuilder()
        # When
        actual = value_builder.bind()
        # Then
        self.assertEqual(Absent.get_absent(), actual)



#  Copyright 2015-2019 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import asyncio
import unittest
from asyncio import Future
from concurrent import futures
from concurrent.futures.thread import ThreadPoolExecutor
from threading import Thread
from unittest.mock import patch

from aiounittest import async_test
from swimai.client.downlinks import ValueDownlinkView
from swimai.structures import Text
from test.utils import MockWebsocketConnect, MockWebsocket, MockAsyncFunction, MockScheduleTask, MockRaiseException, \
    mock_exception_callback
from swimai import SwimClient


class TestSwimClient(unittest.TestCase):

    def setUp(self):
        MockWebsocket.clear()
        MockScheduleTask.clear()
        asyncio.get_event_loop().stop()
        asyncio.get_event_loop().close()

    def test_swim_client_start(self):
        # Given
        client = SwimClient()
        # When
        actual = client.start()
        # Then
        self.assertEqual(actual, client)
        self.assertIsInstance(actual.loop, asyncio.events.AbstractEventLoop)
        self.assertIsInstance(actual.loop_thread, Thread)
        self.assertFalse(actual.loop.is_closed())
        self.assertTrue(actual.loop_thread.is_alive())
        self.assertIsNone(actual.executor)
        client.stop()

    def test_swim_client_stop(self):
        # Given
        client = SwimClient()
        client.start()
        # When
        actual = client.stop()
        # Then
        self.assertEqual(client, actual)
        self.assertIsInstance(actual.loop, asyncio.events.AbstractEventLoop)
        self.assertIsInstance(actual.loop_thread, Thread)
        self.assertTrue(actual.loop.is_closed())
        self.assertFalse(actual.loop_thread.is_alive())
        self.assertIsNone(actual.executor)

    def test_swim_client_with_statement(self):
        # When
        with SwimClient() as swim_client:
            # Then
            self.assertIsInstance(swim_client, SwimClient)
            self.assertIsInstance(swim_client.loop, asyncio.events.AbstractEventLoop)
            self.assertIsInstance(swim_client.loop_thread, Thread)
            self.assertFalse(swim_client.loop.is_closed())

        self.assertIsInstance(swim_client.loop, asyncio.events.AbstractEventLoop)
        self.assertIsInstance(swim_client.loop_thread, Thread)
        self.assertTrue(swim_client.loop.is_closed())
        self.assertFalse(swim_client.loop_thread.is_alive())
        self.assertIsNone(swim_client.executor)

    @patch('warnings.warn')
    @patch('traceback.print_tb')
    def test_swim_client_with_statement_exception_no_handler(self, mock_warn_tb, mock_warn):
        # When
        with SwimClient(debug=True) as swim_client:
            # Then
            self.assertIsInstance(swim_client, SwimClient)
            self.assertIsInstance(swim_client.loop, asyncio.events.AbstractEventLoop)
            self.assertIsInstance(swim_client.loop_thread, Thread)
            self.assertFalse(swim_client.loop.is_closed())
            self.assertTrue(swim_client.loop_thread.is_alive())

            swim_client.task_with_exception = lambda: (_ for _ in ()).throw(Exception,
                                                                            Exception('Mock exception in task'))

            swim_client.task_with_exception()

        mock_warn.assert_called_once()
        mock_warn_tb.assert_called_once()
        self.assertEqual('Mock exception in task', mock_warn.call_args_list[0][0][0].args[0])
        self.assertTrue(swim_client.loop.is_closed())
        self.assertFalse(swim_client.loop_thread.is_alive())
        self.assertIsNone(swim_client.executor)

    @patch('warnings.warn')
    @patch('traceback.print_tb')
    @patch('os._exit')
    def test_swim_client_with_statement_exception_terminate(self, mock_exit, mock_warn_tb, mock_warn):
        # When
        with SwimClient(terminate_on_exception=True, debug=True) as swim_client:
            # Then
            self.assertIsInstance(swim_client, SwimClient)
            self.assertIsInstance(swim_client.loop, asyncio.events.AbstractEventLoop)
            self.assertFalse(swim_client.loop.is_closed())
            self.assertIsInstance(swim_client.loop_thread, Thread)
            self.assertTrue(swim_client.loop_thread.is_alive())

            swim_client.task_with_exception = lambda: (_ for _ in ()).throw(Exception,
                                                                            Exception('Mock exception in task'))

            swim_client.task_with_exception()

        self.assertTrue(swim_client.loop.is_closed())
        self.assertEqual('Mock exception in task', mock_warn.call_args_list[0][0][0].args[0])
        mock_warn_tb.assert_called_once()
        mock_warn.assert_called_once()
        mock_exit.assert_called_once_with(1)
        self.assertFalse(swim_client.loop_thread.is_alive())
        self.assertIsNone(swim_client.executor)

    @patch('warnings.warn')
    @patch('traceback.print_tb')
    @patch('builtins.print')
    def test_swim_client_with_statement_exception_callback(self, mock_print, mock_warn_tb, mock_warn):
        # Given
        mock_callback = mock_exception_callback
        # When
        with SwimClient(execute_on_exception=mock_callback, debug=True) as swim_client:
            # Then
            self.assertIsInstance(swim_client.loop, asyncio.events.AbstractEventLoop)
            self.assertTrue(swim_client.loop_thread.is_alive())
            self.assertIsInstance(swim_client, SwimClient)
            self.assertFalse(swim_client.loop.is_closed())
            self.assertIsInstance(swim_client.loop_thread, Thread)

            swim_client.task_with_exception = lambda: (_ for _ in ()).throw(Exception,
                                                                            Exception('Mock exception in task'))

            swim_client.task_with_exception()

        self.assertEqual('Mock exception callback', mock_print.call_args_list[0][0][0])
        self.assertEqual('Mock exception in task', mock_warn.call_args_list[0][0][0].args[0])
        self.assertTrue(swim_client.loop.is_closed())
        mock_warn_tb.assert_called_once()
        self.assertFalse(swim_client.loop_thread.is_alive())

    @patch('warnings.warn')
    @patch('traceback.print_tb')
    @patch('os._exit')
    def test_swim_client_with_statement_exception_callback_and_terminate(self, mock_exit, mock_warn_tb, mock_warn):
        # Given
        mock_callback = mock_exception_callback
        # When
        with SwimClient(terminate_on_exception=True, execute_on_exception=mock_callback, debug=True) as swim_client:
            # Then
            self.assertIsInstance(swim_client, SwimClient)
            self.assertIsInstance(swim_client.loop, asyncio.events.AbstractEventLoop)
            self.assertFalse(swim_client.loop.is_closed())
            self.assertTrue(swim_client.loop_thread.is_alive())
            self.assertIsInstance(swim_client.loop_thread, Thread)

            swim_client.task_with_exception = lambda: (_ for _ in ()).throw(Exception,
                                                                            Exception('Mock exception in task'))

            swim_client.task_with_exception()

        mock_exit.assert_called_once_with(1)
        self.assertTrue(swim_client.loop.is_closed())
        self.assertEqual('Mock exception in task', mock_warn.call_args_list[0][0][0].args[0])
        mock_warn_tb.assert_called_once()
        mock_warn.assert_called_once()
        self.assertIsNone(swim_client.executor)
        self.assertFalse(swim_client.loop_thread.is_alive())

    @patch('websockets.connect', new_callable=MockWebsocketConnect)
    def test_swim_client_command(self, mock_websocket_connect):
        # Given
        host_uri = 'ws://localhost:9001'
        node_uri = 'moo'
        lane_uri = 'cow'
        expected = '@command(node:moo,lane:cow)"Hello, World!"'
        with SwimClient() as swim_client:
            # When
            actual = swim_client.command(host_uri, node_uri, lane_uri, Text.create_from('Hello, World!'))
            while not actual.done():
                pass

        # Then
        self.assertIsInstance(actual, futures.Future)
        mock_websocket_connect.assert_called_once_with(host_uri)
        self.assertEqual(expected, MockWebsocket.get_mock_websocket().sent_messages[0])

    def test_swim_client_downlink_value(self):
        # Given
        with SwimClient() as swim_client:
            # When
            downlink_view = swim_client.downlink_value()

        # Then
        self.assertIsInstance(downlink_view, ValueDownlinkView)
        self.assertEqual(downlink_view.client, swim_client)

    @patch('swimai.client.connections.ConnectionPool.add_downlink_view', new_callable=MockAsyncFunction)
    @async_test
    async def test_swim_client_add_downlink_view(self, mock_add_downlink):
        # Given
        host_uri = 'ws://localhost:9001'
        node_uri = 'moo'
        lane_uri = 'cow'

        MockWebsocket.get_mock_websocket().messages_to_send.append('@synced(node:"moo",lane:"cow")')

        with SwimClient() as swim_client:
            downlink_view = swim_client.downlink_value()
            downlink_view.host_uri = host_uri
            downlink_view.node_uri = node_uri
            downlink_view.lane_uri = lane_uri
            # When
            await swim_client.add_downlink_view(downlink_view)

        # Then
        mock_add_downlink.assert_called_once_with(downlink_view)

    @patch('swimai.client.connections.ConnectionPool.add_downlink_view', new_callable=MockAsyncFunction)
    @patch('swimai.client.connections.ConnectionPool.remove_downlink_view', new_callable=MockAsyncFunction)
    @async_test
    async def test_swim_client_remove_downlink_view(self, mock_remove_downlink, mock_add_downlink):
        # Given
        host_uri = 'ws://localhost:9001'
        node_uri = 'moo'
        lane_uri = 'cow'

        MockWebsocket.get_mock_websocket().messages_to_send.append('@synced(node:"moo",lane:"cow")')

        with SwimClient() as swim_client:
            downlink_view = swim_client.downlink_value()
            downlink_view.host_uri = host_uri
            downlink_view.node_uri = node_uri
            downlink_view.lane_uri = lane_uri
            await swim_client.add_downlink_view(downlink_view)
            # When
            await swim_client.remove_downlink_view(downlink_view)

        # Then
        mock_add_downlink.assert_called_once_with(downlink_view)
        mock_remove_downlink.assert_called_once_with(downlink_view)

    @patch('swimai.client.connections.ConnectionPool.get_connection', new_callable=MockAsyncFunction)
    @async_test
    async def test_swim_client_get_connection(self, mock_get_connection):
        # Given
        host_uri = 'ws://localhost:9001'
        node_uri = 'moo'
        lane_uri = 'cow'

        MockWebsocket.get_mock_websocket().messages_to_send.append('@synced(node:"moo",lane:"cow")')

        with SwimClient() as swim_client:
            downlink_view = swim_client.downlink_value()
            downlink_view.host_uri = host_uri
            downlink_view.node_uri = node_uri
            downlink_view.lane_uri = lane_uri
            # When
            await swim_client.get_connection(host_uri)

        # Then
        mock_get_connection.assert_called_once_with(host_uri)

    @async_test
    async def test_swim_client_test_schedule_task_async(self):
        #  Given
        mock_task = MockScheduleTask.get_mock_schedule_task()
        with SwimClient() as swim_client:
            # When
            actual = swim_client.schedule_task(mock_task.async_execute, 'foo')
            while not actual.done():
                pass

        # Then
        self.assertEqual(1, mock_task.call_count)
        self.assertEqual('foo', mock_task.message)
        self.assertIsInstance(actual, futures.Future)

    def test_swim_client_test_schedule_task_sync(self):
        # Given
        mock_task = MockScheduleTask.get_mock_schedule_task()
        with SwimClient() as swim_client:
            # When
            actual = swim_client.schedule_task(mock_task.sync_execute, 'foo')
            while not actual.done():
                pass

        # Then
        self.assertEqual(1, mock_task.call_count)
        self.assertEqual('foo', mock_task.message)
        self.assertIsInstance(actual, Future)
        self.assertIsInstance(swim_client.executor, ThreadPoolExecutor)

    @patch('warnings.warn')
    @patch('traceback.print_tb')
    @patch('asyncio.base_events.BaseEventLoop.run_in_executor', new_callable=MockRaiseException)
    def test_swim_client_test_schedule_task_exception(self, mock_run_in_executor, mock_warn_tb, mock_warn):
        # Given
        mock_task = MockScheduleTask.get_mock_schedule_task()
        # When
        with SwimClient() as swim_client:
            actual = swim_client.schedule_task(mock_task.sync_execute, 'foo')

        self.assertEqual('Mock exception', mock_warn.call_args_list[0][0][0].args[0])
        mock_warn_tb.assert_not_called()
        self.assertIsNone(actual)
        mock_run_in_executor.assert_called_once()

    @patch('warnings.warn')
    @patch('traceback.print_tb')
    @patch('asyncio.base_events.BaseEventLoop.run_in_executor', new_callable=MockRaiseException)
    def test_swim_client_test_schedule_task_exception_debug(self, mock_run_in_executor, mock_warn_tb, mock_warn):
        # Given
        mock_task = MockScheduleTask.get_mock_schedule_task()
        # When
        with SwimClient(debug=True) as swim_client:
            actual = swim_client.schedule_task(mock_task.sync_execute, 'foo')

        self.assertEqual('Mock exception', mock_warn.call_args_list[0][0][0].args[0])
        mock_warn_tb.assert_called_once()
        self.assertIsNone(actual)
        mock_run_in_executor.assert_called_once()

    @patch('warnings.warn')
    @patch('traceback.print_tb')
    def test_swim_client_test_schedule_async_task_that_raises_exception(self, mock_warn_tb, mock_warn):
        # Given
        mock_task = MockScheduleTask.get_mock_schedule_task()
        with SwimClient(debug=True) as swim_client:
            # When
            actual = swim_client.schedule_task(mock_task.async_exception_execute, 'foo')
            while not actual.done():
                pass

        # Then
        mock_warn_tb.assert_called_once()
        mock_warn.assert_called_once()
        self.assertEqual('Mock async execute exception', mock_warn.call_args_list[0][0][0].args[0])
        self.assertEqual(1, mock_task.call_count)
        self.assertEqual('foo', mock_task.message)
        self.assertIsInstance(actual, futures.Future)

    @patch('warnings.warn')
    @patch('traceback.print_tb')
    def test_swim_client_test_schedule_sync_task_that_raises_exception(self, mock_warn_tb, mock_warn):
        # Given
        mock_task = MockScheduleTask.get_mock_schedule_task()
        with SwimClient(debug=True) as swim_client:
            # When
            actual = swim_client.schedule_task(mock_task.sync_exception_execute, 'foo')
            while not actual.done():
                pass

        # Then
        mock_warn_tb.assert_called_once()
        mock_warn.assert_called_once()
        self.assertEqual(1, mock_task.call_count)
        self.assertEqual('Mock sync execute exception', mock_warn.call_args_list[0][0][0].args[0])
        self.assertEqual('foo', mock_task.message)
        self.assertIsInstance(actual, Future)

    def test_swim_client_test_schedule_async_task_that_is_cancelled(self):
        # Given
        mock_task = MockScheduleTask.get_mock_schedule_task()
        with SwimClient() as swim_client:
            # When
            actual = swim_client.schedule_task(mock_task.async_infinite_cancel_execute)
        # Then
        self.assertIsInstance(actual, futures.Future)
        self.assertTrue(actual.cancelled())



#  Copyright 2015-2019 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import math
from abc import ABC, abstractmethod
from typing import Optional

from swimai.recon import Recon
from swimai.structures import Item, Record, Attr, Value, Num, RecordMap


class Envelope(ABC):

    def __init__(self, node_uri: str, lane_uri: str, tag: str, form: 'Form', body: Item = Value.absent()) -> None:
        self.node_uri = node_uri
        self.lane_uri = lane_uri
        self.tag = tag
        self.form = form
        self.body = body

    @property
    def route(self):
        return f'{self.node_uri}/{self.lane_uri}'

    @staticmethod
    async def parse_recon(recon_message: str) -> Optional['Envelope']:
        """
        Parse a Recon message in string format into an Envelope.

        :param recon_message    - Recon message in string format.
        :return:                - Envelope from the Recon message.
        """
        value = await Recon.parse(recon_message)
        if isinstance(value, RecordMap):
            return Envelope.create_from_value(value)

    @staticmethod
    def create_from_value(value: RecordMap) -> 'Envelope':
        """
        Parse a Swim value object into an Envelope.

        :param value:           - Swim value object.
        :return:                - Envelope from the Swim value object.
        """
        tag = value.tag
        form = Envelope.resolve_form(tag)

        return form.cast(value)

    @staticmethod
    def resolve_form(tag: str) -> 'Form':
        """
        Returns a Swim form corresponding to a given tag.

        :param tag:             - Name of the tag as string.
        :return:                - The form corresponding to the tag.
        """
        if tag == 'sync':
            return SyncRequestForm()
        if tag == 'synced':
            return SyncedResponseForm()
        if tag == 'linked':
            return LinkedResponseForm()
        if tag == 'event':
            return EventMessageForm()
        if tag == 'command':
            return CommandMessageForm()
        else:
            raise TypeError(f'Invalid form tag: {tag}')

    async def to_recon(self) -> str:
        """
        Create a Recon message in string format representing this Envelope.

        :return:                - Recon message in string format from this Envelope.
        """
        return await Recon.to_string(self.to_value())

    def to_value(self) -> Value:
        """
        Create a Swim value object representing this Envelope.

        :return:                - Swim value object from this Envelope.
        """
        return self.form.mold(self)


class LinkAddressedEnvelope(Envelope):

    def __init__(self, node_uri: str, lane_uri: str, prio: float, rate: float, tag: str, form: 'Form',
                 body: Item = Value.absent()) -> None:
        super().__init__(node_uri, lane_uri, tag, form, body)
        self.prio = prio
        self.rate = rate


class LaneAddressedEnvelope(Envelope):

    def __init__(self, node_uri: str, lane_uri: str, tag: str, form: 'Form', body=Value.absent()) -> None:
        super().__init__(node_uri, lane_uri, tag, form, body)


class SyncRequest(LinkAddressedEnvelope):

    def __init__(self, node_uri: str, lane_uri: str, prio: float = 0.0, rate: float = 0.0,
                 body: Item = Value.absent()) -> None:
        super().__init__(node_uri, lane_uri, prio, rate, tag='sync', form=SyncRequestForm(), body=body)


class LinkedResponse(LinkAddressedEnvelope):

    def __init__(self, node_uri: str, lane_uri: str, prio: float = 0.0, rate: float = 0.0,
                 body: Item = Value.absent()) -> None:
        super().__init__(node_uri, lane_uri, prio, rate, tag='linked', form=LinkedResponseForm(), body=body)


class SyncedResponse(LaneAddressedEnvelope):

    def __init__(self, node_uri: str, lane_uri: str, body: Item = Value.absent()) -> None:
        super().__init__(node_uri, lane_uri, tag='synced', form=SyncedResponseForm(), body=body)


class CommandMessage(LaneAddressedEnvelope):

    def __init__(self, node_uri: str, lane_uri: str, body: Item = Value.absent()) -> None:
        super().__init__(node_uri, lane_uri, tag='command', form=CommandMessageForm(), body=body)


class EventMessage(LaneAddressedEnvelope):

    def __init__(self, node_uri: str, lane_uri: str, body: Item = Value.absent()) -> None:
        super().__init__(node_uri, lane_uri, tag='event', form=EventMessageForm(), body=body)


class Form(ABC):

    @property
    @abstractmethod
    def tag(self) -> str:
        """
        Return the tag associated with the given Form object.

        :return:                - Name of the tag as string value.
        """
        raise NotImplementedError

    @abstractmethod
    def mold(self, envelope: 'Envelope') -> 'Value':
        """
        Create a value object from a given Envelope.

        :param envelope:        - Envelope to convert to Value object.
        :return:                - Value object representing the Envelope.
        """
        raise NotImplementedError

    @abstractmethod
    def cast(self, item: RecordMap) -> 'Envelope':
        """
        Create an Envelope object from a RecordMap.

        :param item:            - RecordMap to convert to Envelope.
        :return:                - Envelope object created from the RecordMap.
        """
        raise NotImplementedError


class LinkAddressedForm(Form):

    @abstractmethod
    def create_envelope_from(self, node_uri: str, lane_uri: str, prio: float, rate: float, body: 'Item') -> 'Envelope':
        """
        Create an Envelope object corresponding to the given LinkAddressedForm.

        :param node_uri:        - Node URI of the Envelope.
        :param lane_uri:        - Lane URI of the Envelope.
        :param prio:            - Priority of the Envelope.
        :param rate:            - Rate of the Envelope.
        :param body:            - Body of the Envelope.
        :return:                - Envelope corresponding to the LinkAddressedForm.
        """
        raise NotImplementedError

    def mold(self, envelope: Optional['LinkAddressedEnvelope']) -> 'Value':

        if envelope is not None:

            headers = Record.create().add_slot('node', envelope.node_uri).add_slot('lane', envelope.lane_uri)
            prio = envelope.prio

            if prio != 0 and not math.isnan(prio):
                headers.add_slot('prio', Num.create_from(prio))

            rate = envelope.rate

            if rate != 0 and not math.isnan(rate):
                headers.add_slot('rate', Num.create_from(rate))

            return Attr.create_attr(self.tag, headers).concat(envelope.body)
        else:
            return Item.extant()

    def cast(self, value: RecordMap) -> Optional['Envelope']:
        headers = value.get_headers(self.tag)
        node_uri = None
        lane_uri = None
        prio = 0.0
        rate = 0.0

        for index in range(0, headers.size):
            header = headers.get_item(index)
            key = header.key.get_string_value()

            if key is not None:

                if key == 'node':
                    node_uri = header.value.get_string_value()
                elif key == 'lane':
                    lane_uri = header.value.get_string_value()
                elif key == 'prio':
                    prio = header.value.get_num_value()
                elif key == 'rate':
                    rate = header.value.get_num_value()

        if node_uri is not None and lane_uri is not None:
            body = value.get_body()
            return self.create_envelope_from(node_uri, lane_uri, prio, rate, body)


class LaneAddressedForm(Form):

    @abstractmethod
    def create_envelope_from(self, node_uri, lane_uri, body) -> 'Envelope':
        """
        :param node_uri:        - Node URI of the Envelope.
        :param lane_uri:        - Lane URI of the Envelope.
        :param body:            - Body of the Envelope.
        :return:                - Envelope corresponding to the LaneAddressedForm.
        """
        raise NotImplementedError

    def mold(self, envelope: Optional['LaneAddressedEnvelope']) -> 'Value':

        if envelope is not None:
            headers = Record.create().add_slot('node', envelope.node_uri).add_slot('lane', envelope.lane_uri)
            return Attr.create_attr(self.tag, headers).concat(envelope.body)
        else:
            return Item.extant()

    def cast(self, item: 'RecordMap') -> Optional['Envelope']:
        value = item
        headers = value.get_headers(self.tag)
        node_uri = None
        lane_uri = None

        for index in range(0, headers.size):
            header = headers.get_item(index)
            key = header.key.get_string_value()

            if key is not None:

                if key == 'node':
                    node_uri = header.value.get_string_value()
                elif key == 'lane':
                    lane_uri = header.value.get_string_value()

        if node_uri is not None and lane_uri is not None:
            body = value.get_body()
            return self.create_envelope_from(node_uri, lane_uri, body)


class SyncedResponseForm(LaneAddressedForm):

    @property
    def tag(self) -> str:
        return 'synced'

    def create_envelope_from(self, node_uri, lane_uri, body) -> 'Envelope':
        return SyncedResponse(node_uri, lane_uri, body=body)


class SyncRequestForm(LinkAddressedForm):

    @property
    def tag(self) -> str:
        return 'sync'

    def create_envelope_from(self, node_uri, lane_uri, prio, rate, body) -> 'Envelope':
        return SyncRequest(node_uri, lane_uri, prio, rate, body=body)


class LinkedResponseForm(LinkAddressedForm):

    @property
    def tag(self) -> str:
        return 'linked'

    def create_envelope_from(self, node_uri, lane_uri, prio, rate, body) -> 'Envelope':
        return LinkedResponse(node_uri, lane_uri, prio, rate, body=body)


class CommandMessageForm(LaneAddressedForm):

    @property
    def tag(self) -> str:
        return 'command'

    def create_envelope_from(self, node_uri, lane_uri, body) -> 'Envelope':
        return CommandMessage(node_uri, lane_uri, body=body)


class EventMessageForm(LaneAddressedForm):

    @property
    def tag(self) -> str:
        return 'event'

    def create_envelope_from(self, node_uri, lane_uri, body) -> 'Envelope':
        return EventMessage(node_uri, lane_uri, body)



#  Copyright 2015-2019 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
import inspect

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from ..utils import URI
from swimai.structures import Absent, Value, Attr, Slot, RecordMap, Bool, Num, Text, Extant, Item
from swimai.warp import SyncRequest, CommandMessage, Envelope
from .downlink_utils import before_open

# Imports for type annotations
if TYPE_CHECKING:
    from ..swim_client import SwimClient
    from ..connections import DownlinkManager


class ValueDownlinkModel:

    def __init__(self, client: 'SwimClient') -> None:
        self.client = client
        self.host_uri = None
        self.node_uri = None
        self.lane_uri = None
        self.form = None
        self.connection = None
        self.task = None
        self.downlink_manager = None
        self.value = Value.absent()

        self.linked = asyncio.Event()
        self.synced = asyncio.Event()

    def open(self) -> 'ValueDownlinkModel':
        self.task = self.client.schedule_task(self.connection.wait_for_messages)
        return self

    def close(self) -> 'ValueDownlinkModel':
        self.client.schedule_task(self.__close)
        return self

    async def establish_downlink(self) -> None:
        """
        Send a `sync` request in order to initiate a connection to a lane from the remote agent.
        """
        sync_request = SyncRequest(self.node_uri, self.lane_uri)
        await self.connection.send_message(await sync_request.to_recon())

    async def receive_message(self, message: 'Envelope') -> None:
        """
        Handle a message from the remote agent.

        :param message:         - Message received from the remote agent.
        """
        if message.tag == 'linked':
            self.linked.set()
        elif message.tag == 'synced':
            self.synced.set()
        elif message.tag == 'event':
            await self.__set_value(message)

    async def send_message(self, message: 'Envelope') -> None:
        """
        Send a message to the remote agent of the downlink.

        :param message:         - Message to send to the remote agent.
        """
        await self.linked.wait()
        await self.connection.send_message(await message.to_recon())

    async def get_value(self) -> Any:
        """
        Get the value of the downlink after it has been synced.

        :return:                - The current value of the downlink.
        """
        await self.synced.wait()
        return self.value

    async def __set_value(self, message: 'Envelope') -> None:
        """
        Set the value of the the downlink and trigger the `did_set` callback of the downlink subscribers.

        :param message:        - The message from the remote agent.
        :return:
        """
        old_value = self.value

        if message.body == Absent.get_absent():
            self.value = None
        elif isinstance(message.body, (Text, Num, Bool)):
            self.value = message.body
        else:
            self.value = await self.record_to_object(message.body)

        await self.downlink_manager.subscribers_did_set(self.value, old_value)

    async def record_to_object(self, body):
        new_object = None

        for item in body.get_items():
            if isinstance(item, Attr):
                class_name = item.key.value
                class_object = self.downlink_manager.registered_classes.get(class_name)

                if class_object is not None:
                    new_object = class_object()
                elif not self.downlink_manager.strict:
                    new_object = type(str(class_name), (object,), {})
                else:
                    raise Exception(f'Missing class for {class_name}')

            if isinstance(item, Slot):

                if isinstance(item.value, RecordMap):
                    setattr(new_object, item.key.value, await self.record_to_object(item.value))
                else:
                    setattr(new_object, item.key.value, item.value.value)

        return new_object

    async def object_to_record(self, obj):

        if isinstance(obj, (str, float, int, bool)):
            recon = Value.create_from(obj)
        else:
            recon = RecordMap.create()
            attr_value = Text.create_from(obj.__class__.__name__)
            recon.add(Attr.create_attr(attr_value, Extant.get_extant()))

            for key, value in obj.__dict__.items():
                slot_value = await self.object_to_record(value)
                key_value = Text.create_from(key)
                recon.add(Slot.create_slot(key_value, slot_value))

        return recon

    async def __close(self) -> None:
        self.task.cancel()


class ValueDownlinkView:

    def __init__(self, client: 'SwimClient') -> None:
        self.client = client
        self.host_uri = None
        self.node_uri = None
        self.lane_uri = None
        self.did_set_callback = None
        self.is_open = False
        self.initialised = asyncio.Event()
        self.model = None
        self.connection = None
        self.__registered_classes = dict()
        self.__deregister_classes = set()
        self.__strict = False
        self.downlink_manager = None

    @property
    def route(self) -> str:
        return f'{self.node_uri}/{self.lane_uri}'

    @property
    def registered_classes(self):
        if self.downlink_manager is None:
            return self.__registered_classes
        else:
            return self.downlink_manager.registered_classes

    @property
    def strict(self):
        if self.downlink_manager is None:
            return self.__strict
        else:
            return self.downlink_manager.strict

    @strict.setter
    def strict(self, strict):
        if self.downlink_manager is not None:
            self.downlink_manager.strict = self.__strict
        else:
            self.__strict = strict

    def open(self) -> 'ValueDownlinkView':

        if not self.is_open:
            self.is_open = True
            self.client.schedule_task(self.client.add_downlink_view, self)

        return self

    def close(self) -> 'ValueDownlinkView':

        if self.is_open:
            self.is_open = False
            self.client.schedule_task(self.client.remove_downlink_view, self)

        return self

    async def create_downlink_model(self, downlink_manager: 'DownlinkManager') -> 'ValueDownlinkModel':
        model = ValueDownlinkModel(self.client)
        downlink_manager.registered_classes = self.__registered_classes
        downlink_manager.strict = self.__strict
        model.downlink_manager = downlink_manager
        model.host_uri = self.host_uri
        model.node_uri = self.node_uri
        model.lane_uri = self.lane_uri
        return model

    @before_open
    def set_host_uri(self, host_uri: str) -> 'ValueDownlinkView':
        self.host_uri = URI.normalise_warp_scheme(host_uri)
        return self

    @before_open
    def set_node_uri(self, node_uri: str) -> 'ValueDownlinkView':
        self.node_uri = node_uri
        return self

    @before_open
    def set_lane_uri(self, lane_uri: str) -> 'ValueDownlinkView':
        self.lane_uri = lane_uri
        return self

    def did_set(self, function: Callable) -> 'ValueDownlinkView':

        if inspect.iscoroutinefunction(function) or isinstance(function, Callable):
            self.did_set_callback = function
        else:
            raise TypeError('Callback must be a coroutine or function!')

        return self

    def get(self, wait_sync: bool = False) -> Any:
        """
        Return the value of the downlink.

        :param wait_sync:       - If True, wait for the initial `sync` to be completed before returning.
                                  If False, return immediately.
        :return:                - The value of the Downlink.
        """
        if self.is_open:
            if wait_sync:
                task = self.client.schedule_task(self.model.get_value)
                return task.result()
            else:
                return self.model.value
        else:
            raise RuntimeError('Link is not open!')

    def set(self, value: Any) -> None:
        """
        Send a command message to set the value of the lane on the remote agent to the given value.

        :param value:           - New value for the lane of the remote agent.
        """
        if self.is_open:
            self.client.schedule_task(self.send_message, value)
        else:
            raise RuntimeError('Link is not open!')

    # noinspection PyAsyncCall
    async def execute_did_set(self, current_value: Any, old_value: Any) -> None:
        """
        Execute the custom `did_set` callback of the current downlink view.

        :param current_value:       - The new value of the downlink.
        :param old_value:           - The previous value of the downlink.
        """
        if self.did_set_callback:
            self.client.schedule_task(self.did_set_callback, current_value, old_value)

    async def send_message(self, value: Any) -> None:
        """
        Send a message to the remote agent of the downlink.

        :param value:           - New value for the lane of the remote agent.
        """
        await self.initialised.wait()

        if isinstance(value, Item):
            recon = value
        else:
            recon = await self.model.object_to_record(value)

        message = CommandMessage(self.node_uri, self.lane_uri, recon)

        await self.model.send_message(message)

    def register_classes(self, classes_list: list) -> None:
        for custom_class in classes_list:
            self.client.schedule_task(self.__register_class, custom_class)

    def register_class(self, custom_class: Any) -> None:
        self.client.schedule_task(self.__register_class, custom_class)

    def deregister_all_classes(self):
        if self.downlink_manager is not None:
            self.__deregister_classes.update(set(self.downlink_manager.registered_classes.keys()))
            self.downlink_manager.registered_classes.clear()
        else:
            self.__registered_classes.clear()

    def deregister_classes(self, classes_list: list) -> None:
        for custom_class in classes_list:
            self.deregister_class(custom_class)

    def deregister_class(self, custom_class: Any) -> None:
        if self.downlink_manager is not None:
            self.downlink_manager.registered_classes.pop(custom_class.__name__, None)
        else:
            self.__registered_classes.pop(custom_class.__name__, None)
            self.__deregister_classes.add(custom_class.__name__)

    def __register_class(self, custom_class: Any) -> None:
        try:
            custom_class()

            if self.downlink_manager is not None:
                self.downlink_manager.registered_classes[custom_class.__name__] = custom_class
            else:
                self.__registered_classes[custom_class.__name__] = custom_class
                self.__deregister_classes.discard(custom_class.__name__)

        except Exception:
            raise Exception(
                f'Class {custom_class.__name__} must have a default constructor or default values for all arguments!')



#  Copyright 2015-2019 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from swimai.structures.structs import Value
from .parsers import ReconParser
from .writers import ReconWriter


class Recon:
    # Singletons
    writer = None
    parser = None

    @staticmethod
    async def parse(recon_string: str) -> 'Value':
        """
        Parse a Recon message in string format and return a Swim structure object.

        :param recon_string:        - Recon message in string format.
        :return:                    - Swim structure object representing the Recon message.
        """
        return await Recon.get_parser().parse_block_string(recon_string)

    @staticmethod
    async def to_string(item: 'Value') -> str:
        """
        Parse a Swim structure object to a Recon string.

        :param item:               - Swim structure object.
        :return:                   - Recon message in string format representing the Swim structure object.
        """
        return await Recon.get_writer().write_item(item)

    @staticmethod
    def get_writer() -> 'ReconWriter':
        """
        Get a Recon writer if one already exists.
        Otherwise, instantiate a new one.

        :return:        - Recon writer.
        """
        if Recon.writer is None:
            Recon.writer = ReconWriter()

        return Recon.writer

    @staticmethod
    def get_parser() -> 'ReconParser':
        """
        Get a Recon parser if one already exists.
        Otherwise, instantiate a new one.

        :return:        - Recon parser.
        """
        if Recon.parser is None:
            Recon.parser = ReconParser()

        return Recon.parser



#  Copyright 2015-2019 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from typing import Union, List, Optional

from .utils import ReconUtils, OutputMessage
from swimai.structures import Attr, Slot, Value, Record, Text, Absent, Num, Extant, Bool, Item


class ReconWriter:

    @staticmethod
    async def write_text(value: str) -> 'OutputMessage':
        if await ReconUtils.is_ident(value):
            return await IdentWriter.write(value=value)
        else:
            return await StringWriter.write(value=value)

    @staticmethod
    async def write_number(value: Union[int, float]) -> 'OutputMessage':
        return await NumberWriter.write(value=value)

    @staticmethod
    async def write_bool(value: bool) -> 'OutputMessage':
        return await BoolWriter.write(value=value)

    @staticmethod
    async def write_absent() -> 'OutputMessage':
        return await OutputMessage.create()

    async def write_item(self, item: 'Item') -> 'str':

        if isinstance(item, Attr):
            output = await self.write_attr(item.key, item.value)
            return output.message
        elif isinstance(item, Slot):
            output = await self.write_slot(item.key, item.value)
            return output.message
        elif isinstance(item, Value):
            output = await self.write_value(item)
            return output.message

        raise TypeError(f'No Recon serialization for {type(item).__name__}!')

    async def write_attr(self, key: 'Value', value: 'Value') -> 'OutputMessage':
        return await AttrWriter.write(key=key, writer=self, value=value)

    async def write_slot(self, key: 'Value', value: 'Value') -> 'OutputMessage':
        return await SlotWriter.write(key=key, writer=self, value=value)

    async def write_value(self, value: Value) -> 'OutputMessage':
        if isinstance(value, Record):
            return await self.write_record(value)
        elif isinstance(value, Text):
            return await self.write_text(value.get_string_value())
        elif isinstance(value, Num):
            return await self.write_number(value.get_num_value())
        elif isinstance(value, Bool):
            return await self.write_bool(value.get_bool_value())
        elif isinstance(value, Absent):
            return await self.write_absent()

    async def write_record(self, record: 'Record') -> Optional['OutputMessage']:
        if record.size > 0:
            message = await BlockWriter.write(items=record.get_items(), writer=self, first=True)
            return message


class Writer(ABC):
    @staticmethod
    @abstractmethod
    async def write() -> 'OutputMessage':
        """
        Write an Item object into its string representation.

        :return:                - OutputMessage containing the string representation of the Item object.
        """
        raise NotImplementedError


class BlockWriter(Writer):

    @staticmethod
    async def write(items: List[Item] = None, writer: 'ReconWriter' = None, first: 'bool' = False,
                    in_braces: bool = False) -> 'OutputMessage':
        output = await OutputMessage.create()

        for item in items:

            if isinstance(item, Attr):
                item_text = await writer.write_item(item)
                await output.append(item_text)
            elif isinstance(item, Value) and not isinstance(item, Record):
                item_text = await writer.write_item(item)
                await output.append(item_text)
            else:
                if not first:
                    await output.append(',')
                elif isinstance(item, Slot):
                    if output.size > 0 and output.last_char != '(':
                        await output.append('{')
                        in_braces = True

                item_text = await writer.write_item(item)
                await output.append(item_text)
                first = False

        if in_braces:
            await output.append('}')

        return output


class AttrWriter(Writer):

    @staticmethod
    async def write(key: 'Value' = None, writer: 'ReconWriter' = None, value: 'Value' = None) -> 'OutputMessage':

        output = await OutputMessage.create('@')
        key_text = await writer.write_value(key)

        if key_text:
            await output.append(key_text)

        if value != Extant.get_extant() and value is not None:
            await output.append('(')
            value_text = await writer.write_value(value)
            await output.append(value_text)
            await output.append(')')

        return output


class SlotWriter(Writer):

    @staticmethod
    async def write(key: Value = None, writer: 'ReconWriter' = None, value: 'Value' = None) -> 'OutputMessage':

        output = await OutputMessage.create()
        key_text = await writer.write_value(key)

        if key_text:
            await output.append(key_text)

        await output.append(':')
        value_text = await writer.write_value(value)

        if value_text:
            await output.append(value_text)

        return output


class StringWriter(Writer):

    @staticmethod
    async def write(value: str = None) -> 'OutputMessage':
        output = await OutputMessage.create('"')

        if value:
            await output.append(value)

        await output.append('"')

        return output


class NumberWriter(Writer):

    @staticmethod
    async def write(value: Union[int, float] = None) -> 'OutputMessage':
        output = await OutputMessage().create()

        if value is not None:
            await output.append(value)

        return output


class BoolWriter(Writer):

    @staticmethod
    async def write(value: bool = None) -> 'OutputMessage':

        if value:
            return await OutputMessage.create('true')
        else:
            return await OutputMessage.create('false')


class IdentWriter(Writer):

    @staticmethod
    async def write(value: str = None) -> 'OutputMessage':
        output = await OutputMessage.create()

        if value:
            await output.append(value)

        return output



#  Copyright 2015-2019 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from urllib.parse import urlparse, ParseResult


class URI:

    @staticmethod
    def normalise_warp_scheme(uri: str) -> str:
        """
        Normalise all different representations of the WARP scheme to a websocket connection.

        :param uri:             - URI to normalise.
        :return:                - URI with `ws` scheme.
        """
        uri = urlparse(uri)
        if URI.has_valid_scheme(uri):
            uri = uri._replace(scheme='ws')
            return uri.geturl()
        else:
            raise TypeError(f'Invalid scheme "{uri.scheme}" for Warp URI!')

    @staticmethod
    def has_valid_scheme(uri: ParseResult) -> bool:
        """
        Check if a URI has a WARP compatible scheme.

        :param uri:             - URI to check.
        :return:                - True if the given URI has a scheme that is WARP compatible. False otherwise.
        """
        scheme = uri.scheme
        return scheme == 'ws' or scheme == 'warp'



#  Copyright 2015-2019 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

from swimai.structures import Absent, Text
from swimai.warp import Envelope, SyncRequestForm, SyncedResponseForm, LinkedResponseForm, EventMessageForm, \
    CommandMessageForm, SyncRequest, SyncedResponse, LinkedResponse, CommandMessage, EventMessage


class TestEnvelopes(unittest.TestCase):

    def test_resolve_form_sync(self):
        # Given
        tag = 'sync'
        # When
        actual = Envelope.resolve_form(tag)
        # Then
        self.assertIsInstance(actual, SyncRequestForm)

    def test_resolve_form_synced(self):
        # Given
        tag = 'synced'
        # When
        actual = Envelope.resolve_form(tag)
        # Then
        self.assertIsInstance(actual, SyncedResponseForm)

    def test_resolve_form_linked(self):
        # Given
        tag = 'linked'
        # When
        actual = Envelope.resolve_form(tag)
        # Then
        self.assertIsInstance(actual, LinkedResponseForm)

    def test_resolve_form_event(self):
        # Given
        tag = 'event'
        # When
        actual = Envelope.resolve_form(tag)
        # Then
        self.assertIsInstance(actual, EventMessageForm)

    def test_resolve_form_command(self):
        # Given
        tag = 'command'
        # When
        actual = Envelope.resolve_form(tag)
        # Then
        self.assertIsInstance(actual, CommandMessageForm)

    def test_resolve_form_invalid(self):
        # Given
        tag = 'this_is_not_a_valid_form'
        # When
        with self.assertRaises(TypeError) as error:
            Envelope.resolve_form(tag)
        # Then
        message = error.exception.args[0]
        self.assertEqual(message, 'Invalid form tag: this_is_not_a_valid_form')

    def test_sync_request_empty_body(self):
        # Given
        node_uri = 'foo__sync_node'
        lane_uri = 'bar__sync_lane'
        # When
        actual = SyncRequest(node_uri, lane_uri)
        # Then
        self.assertEqual('foo__sync_node', actual.node_uri)
        self.assertEqual('bar__sync_lane', actual.lane_uri)
        self.assertEqual(0.0, actual.prio)
        self.assertEqual(0.0, actual.rate)
        self.assertEqual('sync', actual.tag)
        self.assertEqual('foo__sync_node/bar__sync_lane', actual.route)
        self.assertEqual(Absent.get_absent(), actual.body)
        self.assertIsInstance(actual.form, SyncRequestForm)

    def test_sync_request_existing_body(self):
        # Given
        node_uri = 'foo_sync_node'
        lane_uri = 'bar_sync_lane'
        body = Text.create_from('Sync_Body')
        priority = 3.5
        rate = 2.0
        # When
        actual = SyncRequest(node_uri, lane_uri, priority, rate, body=body)
        # Then
        self.assertEqual('foo_sync_node', actual.node_uri)
        self.assertEqual('bar_sync_lane', actual.lane_uri)
        self.assertEqual(3.5, actual.prio)
        self.assertEqual(2.0, actual.rate)
        self.assertEqual('sync', actual.tag)
        self.assertEqual('foo_sync_node/bar_sync_lane', actual.route)
        self.assertEqual(body, actual.body)
        self.assertIsInstance(actual.form, SyncRequestForm)

    def test_synced_response_empty_body(self):
        # Given
        node_uri = 'foo_synced_node'
        lane_uri = 'bar_synced_lane'
        # When
        actual = SyncedResponse(node_uri, lane_uri)
        # Then
        self.assertEqual('foo_synced_node', actual.node_uri)
        self.assertEqual('bar_synced_lane', actual.lane_uri)
        self.assertEqual('synced', actual.tag)
        self.assertEqual('foo_synced_node/bar_synced_lane', actual.route)
        self.assertEqual(Absent.get_absent(), actual.body)
        self.assertIsInstance(actual.form, SyncedResponseForm)

    def test_synced_response_existing_body(self):
        # Given
        node_uri = 'foo_synced_node'
        lane_uri = 'bar_synced_lane'
        body = Text.create_from('Synced_Body')
        # When
        actual = SyncedResponse(node_uri, lane_uri, body=body)
        # Then
        self.assertEqual('foo_synced_node', actual.node_uri)
        self.assertEqual('bar_synced_lane', actual.lane_uri)
        self.assertEqual('synced', actual.tag)
        self.assertEqual('foo_synced_node/bar_synced_lane', actual.route)
        self.assertEqual(body, actual.body)
        self.assertIsInstance(actual.form, SyncedResponseForm)

    def test_linked_response_empty_body(self):
        # Given
        node_uri = 'foo_linked_node'
        lane_uri = 'bar_linked_lane'
        # When
        actual = LinkedResponse(node_uri, lane_uri)
        # Then
        self.assertEqual('foo_linked_node', actual.node_uri)
        self.assertEqual('bar_linked_lane', actual.lane_uri)
        self.assertEqual('linked', actual.tag)
        self.assertEqual('foo_linked_node/bar_linked_lane', actual.route)
        self.assertEqual(Absent.get_absent(), actual.body)
        self.assertIsInstance(actual.form, LinkedResponseForm)

    def test_linked_response_existing_body(self):
        # Given
        node_uri = 'foo_linked_node'
        lane_uri = 'bar_linked_lane'
        body = Text.create_from('Linked_Body')
        priority = 1.13
        rate = 3.11
        # When
        actual = LinkedResponse(node_uri, lane_uri, priority, rate, body)
        # Then
        self.assertEqual('foo_linked_node', actual.node_uri)
        self.assertEqual('bar_linked_lane', actual.lane_uri)
        self.assertEqual('linked', actual.tag)
        self.assertEqual('foo_linked_node/bar_linked_lane', actual.route)
        self.assertEqual(body, actual.body)
        self.assertIsInstance(actual.form, LinkedResponseForm)

    def test_command_message_empty_body(self):
        # Given
        node_uri = 'foo_command_node'
        lane_uri = 'bar_command_lane'
        # When
        actual = CommandMessage(node_uri, lane_uri)
        # Then
        self.assertEqual('foo_command_node', actual.node_uri)
        self.assertEqual('bar_command_lane', actual.lane_uri)
        self.assertEqual('command', actual.tag)
        self.assertEqual('foo_command_node/bar_command_lane', actual.route)
        self.assertEqual(Absent.get_absent(), actual.body)
        self.assertIsInstance(actual.form, CommandMessageForm)

    def test_command_message_existing_body(self):
        # Given
        node_uri = 'foo_command_node'
        lane_uri = 'bar_command_lane'
        body = Text.create_from('Command_Body')
        # When
        actual = CommandMessage(node_uri, lane_uri, body)
        # Then
        self.assertEqual('foo_command_node', actual.node_uri)
        self.assertEqual('bar_command_lane', actual.lane_uri)
        self.assertEqual('command', actual.tag)
        self.assertEqual('foo_command_node/bar_command_lane', actual.route)
        self.assertEqual(body, actual.body)
        self.assertIsInstance(actual.form, CommandMessageForm)

    def test_event_message_empty_body(self):
        # Given
        node_uri = 'foo_event_node'
        lane_uri = 'bar_event_lane'
        # When
        actual = EventMessage(node_uri, lane_uri)
        # Then
        self.assertEqual('foo_event_node', actual.node_uri)
        self.assertEqual('bar_event_lane', actual.lane_uri)
        self.assertEqual('event', actual.tag)
        self.assertEqual('foo_event_node/bar_event_lane', actual.route)
        self.assertEqual(Absent.get_absent(), actual.body)
        self.assertIsInstance(actual.form, EventMessageForm)

    def test_event_message_existing_body(self):
        # Given
        node_uri = 'foo_event_node'
        lane_uri = 'bar_event_lane'
        body = Text.create_from('Event_Body')
        # When
        actual = EventMessage(node_uri, lane_uri, body)
        # Then
        self.assertEqual('foo_event_node', actual.node_uri)
        self.assertEqual('bar_event_lane', actual.lane_uri)
        self.assertEqual('foo_event_node/bar_event_lane', actual.route)
        self.assertEqual('event', actual.tag)
        self.assertEqual(body, actual.body)
        self.assertIsInstance(actual.form, EventMessageForm)



#  Copyright 2015-2019 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from typing import Optional, Union, Any


class ReconUtils:

    @staticmethod
    async def is_ident_start_char(char: Union[str, int]) -> bool:
        """
        Check if a character is a valid first character of an identifier.
        Valid start characters for identifiers: [A-Za-z_]

        :param char:        - Character to check.
        :return:            - True if the character is valid, False otherwise.
        """
        if char:
            char = await ReconUtils.to_ord(char)

            return ord('A') <= char <= ord('Z') or char == ord('_') or ord('a') <= char <= ord('z')
        else:
            return False

    @staticmethod
    async def is_ident_char(char: Union[str, int]) -> bool:
        """
        Check if a character is a valid character of an identifier.
        Valid characters for identifiers: [A-Za-z_-]

        :param char:        - Character to check.
        :return:            - True if the character is valid, False otherwise.
        """
        if char:
            char = await ReconUtils.to_ord(char)
            return char == ord('-') or await ReconUtils.is_digit(char) or await ReconUtils.is_ident_start_char(char)
        else:
            return False

    @staticmethod
    async def is_ident(value: str) -> bool:
        """
        Check if a string value is a valid identifier.

        :param value:      - Value to check.
        :return:           - True if the value is valid identifier, False otherwise.
        """
        if len(value) == 0:
            return False

        if not await ReconUtils.is_ident_start_char(value[0]):
            return False

        for char in value:
            if not await ReconUtils.is_ident_char(char):
                return False

        return True

    @staticmethod
    async def is_space(char: Union[str, int]) -> bool:
        """
        Check if a character is a space character.

        :param char:        - Character to check.
        :return:            - True if the character is a space character, False otherwise.
        """
        if char:
            char = await ReconUtils.to_ord(char)
            return char == ord(' ') or char == ord('\t')
        else:
            return False

    @staticmethod
    async def is_digit(char: Union[str, int]) -> bool:
        """
       Check if a character is a digit.

       :param char:         - Character to check.
       :return:             - True if the character is a digit, False otherwise.
       """
        if char:
            char = await ReconUtils.to_ord(char)
            return ord('0') <= char <= ord('9')
        else:
            return False

    @staticmethod
    async def to_ord(char: Any) -> Optional[int]:
        """
        Convert a character to its integer representation.

        :param char:        - Character to convert.
        :return:            - Integer representation of the character.
        """
        if isinstance(char, str):
            return ord(char)
        if isinstance(char, int):
            return char
        else:
            return None


class Message(ABC):

    def __init__(self) -> None:
        self.message = ''

    @property
    def value(self) -> str:
        return self.message

    @property
    def size(self) -> int:
        return len(self.message)

    @staticmethod
    @abstractmethod
    async def create(chars: str) -> 'Message':
        raise NotImplementedError

    async def append(self, obj: Any) -> None:
        """
        Append the string representation of an object to the current message.

        :param obj:           - Object to append to the message.
        """
        if isinstance(obj, str):
            self.message = self.message + obj
        elif isinstance(obj, (float, int)):
            self.message = self.message + str(obj)
        elif isinstance(obj, (OutputMessage, InputMessage)):
            self.message = self.message + obj.value
        else:
            raise TypeError(f'Item of type {type(obj).__name__} cannot be added to Message!')


class OutputMessage(Message):

    def __init__(self) -> None:
        super().__init__()

    @property
    def last_char(self) -> str:
        """
        Return the last character of the message.

        :return:                - Last character of the message.
        """
        if self.size > 0:
            return self.message[-1]
        else:
            return ''

    @staticmethod
    async def create(chars: str = None) -> 'OutputMessage':
        """
        Create an OutputMessage instance and initialise its message.

        :param chars:           - Initial value of the message.
        :return:                - OutputMessage instance.
        """
        instance = OutputMessage()

        if chars:
            await instance.append(chars)

        return instance


class InputMessage(Message):

    def __init__(self) -> None:
        super().__init__()
        self.index = 0

    @staticmethod
    async def create(chars: str = None) -> 'InputMessage':
        """
        Create an OutputMessage instance and initialise its message.

        :param chars:           - Initial value of the message.
        :return:                - OutputMessage instance.
        """
        instance = InputMessage()

        if chars:
            await instance.append(chars)

        return instance

    @staticmethod
    async def skip_spaces(message: 'InputMessage') -> None:
        """
        Moves the head of the message to the next non-space character.

        :param message:     - InputMessage object.
        """
        char = message.head
        while await ReconUtils.is_space(char):
            char = message.step()

    @property
    def head(self) -> str:
        """
        Get the character at the front of the InputMessage pointed by the message index.

        :return:                - The current head character of the InputMessage.
        """
        if self.is_cont:
            return self.message[self.index]
        else:
            return ''

    @property
    def is_cont(self) -> bool:
        """
        Check if there are any characters left in front of the InputMessage index.

        :return:                - False if the index is pointing at the last character or beyond, True otherwise.
        """
        if self.index >= len(self.message):
            return False
        else:
            return True

    def step(self) -> str:
        """
        Move the head index forward by one.

        :return:                - The new head character of the InputMessage.
        """
        self.index = self.index + 1
        return self.head



#  Copyright 2015-2019 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .swim_client import SwimClient
from .downlinks import ValueDownlinkModel
from .connections import WSConnection, ConnectionStatus, ConnectionPool, DownlinkManagerPool, DownlinkManager, \
    DownlinkManagerStatus
from .utils import URI

__all__ = [SwimClient, ValueDownlinkModel, WSConnection, ConnectionStatus, ConnectionPool, DownlinkManagerPool,
           DownlinkManager, URI, DownlinkManagerStatus]



#  Copyright 2015-2019 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
import unittest

from aiounittest import async_test
from swimai.structures import Num, Text, RecordMap, Attr, Slot, Extant, Bool
from swimai.warp import SyncRequest, SyncedResponse, EventMessage, LinkedResponse, CommandMessage


class TestWriters(unittest.TestCase):

    @async_test
    async def test_write_sync(self):
        # Given
        envelope = SyncRequest('foo', 'bar')
        expected = '@sync(node:foo,lane:bar)'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_sync_escaped(self):
        # Given
        envelope = SyncRequest('bar/baz/', 'foo/bar')
        expected = '@sync(node:"bar/baz/",lane:"foo/bar")'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_sync_body_int(self):
        # Given
        envelope = SyncRequest('bar/baz/', 'foo/bar', body=Num.create_from(42))
        expected = '@sync(node:"bar/baz/",lane:"foo/bar")42'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_sync_body_float(self):
        # Given
        envelope = SyncRequest('bar/baz/', 'foo/bar', body=Num.create_from(33.22))
        expected = '@sync(node:"bar/baz/",lane:"foo/bar")33.22'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_sync_body_bool(self):
        # Given
        envelope = SyncRequest('bar/baz/', 'foo/bar', body=Bool.create_from(False))
        expected = '@sync(node:"bar/baz/",lane:"foo/bar")false'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_sync_body_string(self):
        # Given
        envelope = SyncRequest('bar/baz/', 'foo/bar', body=Text.create_from('Hello, World!'))
        expected = '@sync(node:"bar/baz/",lane:"foo/bar")"Hello, World!"'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_sync_prio(self):
        # Given
        envelope = SyncRequest('bar/baz/', 'foo/bar', prio=33, body=Text.create_from('Hello, World!'))
        expected = '@sync(node:"bar/baz/",lane:"foo/bar",prio:33)"Hello, World!"'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_sync_rate(self):
        # Given
        envelope = SyncRequest('bar/baz/', 'foo/bar', rate=0.3, body=Text.create_from('Hello, World!'))
        expected = '@sync(node:"bar/baz/",lane:"foo/bar",rate:0.3)"Hello, World!"'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_sync_prio_rate(self):
        # Given
        envelope = SyncRequest('bar/baz/', 'foo/bar', prio=33.2, rate=0.3, body=Text.create_from('Hello, World!'))
        expected = '@sync(node:"bar/baz/",lane:"foo/bar",prio:33.2,rate:0.3)"Hello, World!"'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_synced(self):
        # Given
        envelope = SyncedResponse('foo', 'bar')
        expected = '@synced(node:foo,lane:bar)'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_synced_escaped(self):
        # Given
        envelope = SyncedResponse('foo/bar/baz', 'baz/bar/foo')
        expected = '@synced(node:"foo/bar/baz",lane:"baz/bar/foo")'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_synced_body_int(self):
        # Given
        envelope = SyncedResponse('foo/bar/baz', 'baz/bar/foo', body=Num.create_from(-13))
        expected = '@synced(node:"foo/bar/baz",lane:"baz/bar/foo")-13'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_synced_body_float(self):
        # Given
        envelope = SyncedResponse('foo/bar/baz', 'baz/bar/foo', body=Num.create_from(-13.04))
        expected = '@synced(node:"foo/bar/baz",lane:"baz/bar/foo")-13.04'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_synced_body_bool(self):
        # Given
        envelope = SyncedResponse('foo/bar/baz', 'baz/bar/foo', body=Bool.create_from(True))
        expected = '@synced(node:"foo/bar/baz",lane:"baz/bar/foo")true'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_synced_body_string(self):
        # Given
        envelope = SyncedResponse('foo/bar/baz', 'baz/bar/foo', body=Text.create_from('Hello, friend.'))
        expected = '@synced(node:"foo/bar/baz",lane:"baz/bar/foo")"Hello, friend."'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_linked(self):
        # Given
        envelope = LinkedResponse('test', 'foo')
        expected = '@linked(node:test,lane:foo)'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_linked_escaped(self):
        # Given
        envelope = LinkedResponse('/unit/5', '/unit/3')
        expected = '@linked(node:"/unit/5",lane:"/unit/3")'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_linked_body_int(self):
        # Given
        envelope = LinkedResponse('/unit/5', '/unit/3', body=Num.create_from(-100))
        expected = '@linked(node:"/unit/5",lane:"/unit/3")-100'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_linked_body_float(self):
        # Given
        envelope = LinkedResponse('/unit/5', '/unit/3', body=Num.create_from(-100.0))
        expected = '@linked(node:"/unit/5",lane:"/unit/3")-100.0'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_linked_body_bool(self):
        # Given
        envelope = LinkedResponse('/unit/5', '/unit/3', body=Bool.create_from(False))
        expected = '@linked(node:"/unit/5",lane:"/unit/3")false'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_linked_body_string(self):
        # Given
        envelope = LinkedResponse('/unit/5', '/unit/3', body=Text.create_from('spam and \\ham/'))
        expected = '@linked(node:"/unit/5",lane:"/unit/3")"spam and \\ham/"'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_linked_prio(self):
        # Given
        envelope = LinkedResponse('/unit/5', '/unit/3', prio=33.2, body=Text.create_from('spam and \\ham/'))
        expected = '@linked(node:"/unit/5",lane:"/unit/3",prio:33.2)"spam and \\ham/"'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_linked_rate(self):
        # Given
        envelope = LinkedResponse('/unit/5', '/unit/3', rate=0.2, body=Text.create_from('spam and \\ham/'))
        expected = '@linked(node:"/unit/5",lane:"/unit/3",rate:0.2)"spam and \\ham/"'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_linked_prio_rate(self):
        # Given
        envelope = LinkedResponse('/unit/5', '/unit/3', prio=22.11, rate=0.2, body=Text.create_from('spam and \\ham/'))
        expected = '@linked(node:"/unit/5",lane:"/unit/3",prio:22.11,rate:0.2)"spam and \\ham/"'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_command(self):
        # Given
        envelope = CommandMessage('test', 'foo')
        expected = '@command(node:test,lane:foo)'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_command_escaped(self):
        # Given
        envelope = CommandMessage('dead/parrot', 'spam/ham')
        expected = '@command(node:"dead/parrot",lane:"spam/ham")'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_command_body_int(self):
        # Given
        envelope = CommandMessage('dead/parrot', 'spam/ham', body=Num.create_from(911))
        expected = '@command(node:"dead/parrot",lane:"spam/ham")911'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_command_body_float(self):
        # Given
        envelope = CommandMessage('dead/parrot', 'spam/ham', body=Num.create_from(-911.119))
        expected = '@command(node:"dead/parrot",lane:"spam/ham")-911.119'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_command_body_bool(self):
        # Given
        envelope = CommandMessage('dead/parrot', 'spam/ham', body=Bool.create_from(True))
        expected = '@command(node:"dead/parrot",lane:"spam/ham")true'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_command_body_string(self):
        # Given
        envelope = CommandMessage('dead/parrot', 'spam/ham', body=Text.create_from('Polly the parrot.'))
        expected = '@command(node:"dead/parrot",lane:"spam/ham")"Polly the parrot."'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_command_body_remove(self):
        # Given
        envelope = CommandMessage('/unit/foo', 'shoppingCart',
                                  body=RecordMap.create_record_map(
                                      Attr.create_attr(Text.create_from('remove'),
                                                       RecordMap.create_record_map(
                                                           Slot.create_slot(Text.create_from('key'),
                                                                            Text.create_from('FromClientLink'))))))
        expected = '@command(node:"/unit/foo",lane:shoppingCart)@remove(key:FromClientLink)'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_event(self):
        # Given
        envelope = EventMessage('test', 'foo')
        expected = '@event(node:test,lane:foo)'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_event_escaped(self):
        # Given
        envelope = EventMessage('/this/is/spam', 'hello')
        expected = '@event(node:"/this/is/spam",lane:hello)'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_event_body_int(self):
        # Given
        envelope = EventMessage('/this/is/spam', 'hello', body=Num.create_from(1224))
        expected = '@event(node:"/this/is/spam",lane:hello)1224'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_event_body_float(self):
        # Given
        envelope = EventMessage('/this/is/spam', 'hello', body=Num.create_from(33.12))
        expected = '@event(node:"/this/is/spam",lane:hello)33.12'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_event_body_bool(self):
        # Given
        envelope = EventMessage('/this/is/spam', 'hello', body=Bool.create_from(False))
        expected = '@event(node:"/this/is/spam",lane:hello)false'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_event_body_string(self):
        # Given
        envelope = EventMessage('/this/is/spam', 'hello', body=Text.create_from("-33.12 / 6"))
        expected = '@event(node:"/this/is/spam",lane:hello)"-33.12 / 6"'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_event_body_object(self):
        # Given
        body = RecordMap.create()
        body.add(Attr.create_attr(Text.create_from('Person'), Extant.get_extant()))
        body.add(Slot.create_slot(Text.create_from('name'), Text.create_from('Bar')))
        body.add(Slot.create_slot(Text.create_from('age'), Num.create_from(14)))
        body.add(Slot.create_slot(Text.create_from('salary'), Num.create_from(-5.9)))

        envelope = EventMessage('/this/is/spam', 'hello', body=body)
        expected = '@event(node:"/this/is/spam",lane:hello)@Person{name:Bar,age:14,salary:-5.9}'

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)

    @async_test
    async def test_write_event_body_nested(self):
        # Given

        friend = RecordMap.create()
        friend.add(Attr.create_attr(Text.create_from('Person'), Extant.get_extant()))
        friend.add(Slot.create_slot(Text.create_from('name'), Text.create_from('Sam/Spam')))
        friend.add(Slot.create_slot(Text.create_from('age'), Num.create_from(1)))
        friend.add(Slot.create_slot(Text.create_from('salary'), Num.create_from(22)))

        body = RecordMap.create()
        body.add(Attr.create_attr(Text.create_from('Person'), Extant.get_extant()))
        body.add(Slot.create_slot(Text.create_from('name'), Text.create_from('Par')))
        body.add(Slot.create_slot(Text.create_from('age'), Num.create_from(11)))
        body.add(Slot.create_slot(Text.create_from('salary'), Num.create_from(-5.9)))
        body.add(Slot.create_slot(Text.create_from('friend'), friend))

        envelope = EventMessage('/this/is/spam', 'hello', body=body)
        person = '@Person{name:Par,age:11,salary:-5.9,friend:@Person{name:"Sam/Spam",age:1,salary:22}}'
        expected = '@event(node:"/this/is/spam",lane:hello)' + person

        # When
        responses = await asyncio.gather(envelope.to_recon())
        actual = responses[0]

        # Then
        self.assertEqual(expected, actual)



#  Copyright 2015-2019 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
import inspect
import os
import sys
import traceback
import warnings

from asyncio import Future
from concurrent.futures import CancelledError
from threading import Thread
from traceback import TracebackException
from typing import Callable, Any, Optional
from concurrent.futures.thread import ThreadPoolExecutor
from .connections import ConnectionPool, WSConnection
from .downlinks import ValueDownlinkView
from .utils import URI
from swimai.structures import Item
from swimai.warp import CommandMessage


class SwimClient:

    def __init__(self, terminate_on_exception: bool = False, execute_on_exception: Callable = None,
                 debug: bool = False) -> None:
        self.loop = None
        self.loop_thread = None
        self.executor = None
        self.__connection_pool = ConnectionPool()
        self.debug = debug

        self.execute_on_exception = execute_on_exception
        self.terminate_on_exception = terminate_on_exception

    def __enter__(self) -> 'SwimClient':
        self.start()
        return self

    def __exit__(self, exc_type: Optional[type], exc_value: Optional[Exception],
                 exc_traceback: Optional[TracebackException]) -> 'SwimClient':

        if exc_value or exc_traceback:
            self.__handle_exception(exc_value, exc_traceback)

        self.stop()
        return self

    def start(self) -> 'SwimClient':
        """
        Start the Swim client.
        Create a new thread and starts an asyncio loop inside it.
        """
        loop = asyncio.new_event_loop()
        asyncio.get_event_loop_policy().set_event_loop(loop)
        self.loop = loop
        self.loop_thread = Thread(target=self.__start_event_loop)
        self.loop_thread.start()

        return self

    def stop(self) -> 'SwimClient':
        """
        Stop the client.
        Schedule a task for stopping the event loop and its thread and afterwards close the loop.
        """
        self.schedule_task(self.__stop_event_loop)
        self.loop_thread.join()
        self.loop.close()

        return self

    def command(self, host_uri: str, node_uri: str, lane_uri: str, body: 'Item') -> 'Future':
        """
        Send a command message to a command lane on a remote Swim agent.

        :param host_uri:        - Host URI of the remote agent.
        :param node_uri:        - Node URI of the remote agent.
        :param lane_uri:        - Lane URI of the command lane of the remote agent.
        :param body:            - The message body.
        """
        return self.schedule_task(self.__send_command, host_uri, node_uri, lane_uri, body)

    def downlink_value(self) -> 'ValueDownlinkView':
        """
        Create a Value Downlink.
        """
        return ValueDownlinkView(self)

    async def add_downlink_view(self, downlink_view: 'ValueDownlinkView') -> None:
        """
        Add a DownlinkView to the connection pool of the client.

        :param downlink_view:   - DownlinkView to add to the connection pool.
        """
        await self.__connection_pool.add_downlink_view(downlink_view)

    async def remove_downlink_view(self, downlink_view: 'ValueDownlinkView') -> None:
        """
        Remove a DownlinkView from the connection pool of the client.

        :param downlink_view:   - DownlinkView to remove from the connection pool.
        """
        await self.__connection_pool.remove_downlink_view(downlink_view)

    async def get_connection(self, host_uri: str) -> 'WSConnection':
        """
        Get a WebSocket connection to the specified host from the connection pool.

        :param host_uri:        - URI of the host.
        :return:                - WebSocket connection to the host.
        """
        connection = await self.__connection_pool.get_connection(host_uri)
        return connection

    def schedule_task(self, task: Callable, *args: Any) -> 'Future':
        """
        Schedule a task for execution in the asyncio loop.

        :param task:            - Coroutine or function to be executed in the asyncio loop.
        :param args:            - Arguments to be passed to the coroutine or function.
        :return:                - Future object that holds information about the task execution and final result.
        """
        try:
            if inspect.iscoroutinefunction(task):
                future = asyncio.run_coroutine_threadsafe(task(*args), loop=self.loop)
            else:
                future = self.loop.run_in_executor(self.__get_pool_executor(), task, *args)

            future.add_done_callback(self.__exception_handler)
            return future

        except Exception:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            self.__handle_exception(exc_value, exc_traceback)

    def __exception_handler(self, future: Future) -> None:
        """
        Check the result of execution of a future and report any exceptions.

        :param future:          - Future that has been completed.
        """
        try:
            future.result()
        except CancelledError:
            pass
        except Exception:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            self.__handle_exception(exc_value, exc_traceback)

    def __handle_exception(self, exc_value: Optional[Exception], exc_traceback: Optional[TracebackException]) -> None:
        """
        Report exceptions and schedule custom callbacks or client termination, based on the
        Swim Client policies.

        :param exc_value:       - Exception value.
        :param exc_traceback:   - Exception traceback.
        """
        warnings.warn(exc_value)

        if self.debug:
            traceback.print_tb(exc_traceback)

        if self.terminate_on_exception:
            os._exit(1)
            return

        if self.execute_on_exception is not None:
            self.execute_on_exception()

    async def __send_command(self, host_uri: str, node_uri: str, lane_uri: str, body: 'Item') -> None:
        """
        Send a command message to a given host.

        :param host_uri:        - Host URI of the remote agent.
        :param node_uri:        - Node URI of the remote agent.
        :param lane_uri:        - Lane URI of the command lane of the remote agent.
        :param body:            - The message body.
        """
        host_uri = URI.normalise_warp_scheme(host_uri)
        message = CommandMessage(node_uri, lane_uri, body=body)
        connection = await self.get_connection(host_uri)
        await connection.send_message(await message.to_recon())

    def __get_pool_executor(self) -> 'ThreadPoolExecutor':
        """
        Return the singleton Thread pool executor or create one if it does not exist.

        :return:                - Thread pool executor.
        """
        if self.executor is None:
            self.executor = ThreadPoolExecutor()

        return self.executor

    def __start_event_loop(self) -> None:
        asyncio.set_event_loop(self.loop)
        asyncio.get_event_loop().run_forever()

    async def __stop_event_loop(self) -> None:

        if self.executor is not None:
            self.executor.shutdown(wait=False)

        tasks = [task for task in asyncio.all_tasks() if task is not asyncio.current_task()]
        [task.cancel() for task in tasks]
        await asyncio.gather(*tasks, return_exceptions=True)

        self.loop.stop()
